/* ------------------------------------------------------------------------ */
/*                                                                          */
/* [binimg.c]    Partable machine-independant binary format                 */
/*                                                                          */
/* Copyright (c) 1993 by Lars D\olle, Heike Manns                           */
/* ------------------------------------------------------------------------ */

#include "standard.h"
#include "symbols.h"

#include "glo_tab.h"
#include "pathes.h"
#include "hmap.h"
#include "otab.h"
#include "sink.h"

/*!!
  Summary

  [binimg] offers a set of functions to read and write binary files in a portable
  machine-independant format. Within the STYX-system they are used to make date structures
  persistent.

  For each supported data type exist a read and write function with the following signature.

  |   void put<TYPE>(<CTYPE>  x);
  |   void get<TYPE>(<CTYPE> &x);

  Source and target are implicit in these operations. The functions 'getBgn' and 'getEnd'
  open and close a source. To open and close a target one has to use the functions 'putBgn'
  and 'putEnd'. So it is not possible to read or write more than one file at a time.

  Eaxh binary file within the STYX-system starts with a header block. Creation and reading
  e.g. checking is done by the functions 'putHeader' and 'getHeader'.

  The binary files are protected against unauthorized reading and writing by an integrated 
  combined encryption and checking method. Further more they will be compressed.
*/

/*!!
  Files and EOF

  It is not possible to read or write more than one file at a time.

  This module doesn't support an explicit EOF-predicate. It is the responsibility of the user
  to check for EOF. Reading behind EOF causes the program to abort with an error message.
*/

/*D 
  Dies ist kein Schikane, sondern eher eine Entlastung im Vergleich zur
  C-Konvention. In letzterer stellen die von 'fgetc' gelieferten Werte
  eine Vereinigung von 'byte' mit dem Sonderwert 'EOF' dar. Mithin wird
  dort nach jeder Leseoperation eine Fallunterscheidung erzwungen, die
  bei unserer Konvention entfllt - es werden eben nur Datenbytes geliefert.

  Wenn das Dateiende erwartet wird, kann dies durch 'checkEOF' abgesichert
  werden. Dies geschieht automatisch beim Schliessen einer gelesenen Datei.
  (Siehe 'getEnd', weiter unten).
*/

static FILE *f; /* definiert, wenn 'state != stateClosed', s.u. */
static string fid;

static void putbyte(byte x)
/* put byte 'x' to 'f' */
{
  fputc(x,f);
}

static void noIncEvent(float lvl)
/* noop */
{
}

static long SrcFileLen;
static long SrcFileGet;
static void (*IncEvent)(float lvl) = noIncEvent;

void BIN_setIncEvent(void (*evt)(float lvl))
/* set the get-inc event */
{
  IncEvent = evt;
}

static int getbyte(bool eof_ok)
/* get byte 'x' from 'f' */
{ int cc = fgetc(f);
  assert1(eof_ok || cc != EOF,"unexpected end of file '%s'",fid);
  if (cc != EOF)
  {
    SrcFileGet += 1;
    if ((SrcFileGet-1)*100/SrcFileLen != SrcFileGet*100/SrcFileLen)
      (*IncEvent)(SrcFileGet/(float)SrcFileLen);
  }
  return cc;
}

static void checkEOF(void)
/* validates 'EOF' reading */
{
  assert1(fgetc(f) == EOF,"data after logical end of file '%s'",fid);
}

/*D
  Verschlsselungmethode

  Zur Verschlsselung wird ein auf linearer Kongruenz basierendes Verfahren
  verwendete. Beachte, dass sich die Ver- und Entschlsslungsoperation
  subtil darin unterscheiden, dass immer das zuletzt "verschlsselte" Byte
  zur Ermittlung des neuen Schlssels verwendet wird.
*/

#define crypt_1 (52845)
#define crypt_2 (22719)
static word  crypt_R;

static byte Encrypt(byte plain)
/* encrypt a byte */
{
  byte cipher = (plain ^ (crypt_R >> 8));
  crypt_R = (cipher + crypt_R) * crypt_1 + crypt_2;
  return cipher;
}

static byte Decrypt(byte cipher)
/* decrypt a byte */
{
  byte plain = (cipher ^ (crypt_R >> 8));
  crypt_R = (cipher + crypt_R) * crypt_1 + crypt_2;
  return plain;
}

/*D
  Verschlsselung und Validierung

  Der Verschlsselungalgorithmus wird zugleich zur Validierung des
  Dateiinhalts, besonders auch ihrer unverschlsselten Teile verwendet.
  Unter dieser Konvention durchluft der "gesamte" Dateiinhalt das
  Verschlsselungsverfahren. Falls unverschlsselte Ausgabe erfolgt,
  wird lediglich der Seiteneffekt der Verschlsselung, der nach jedem
  Byte einen neuen Schlssel erzeugt verwendet. Diese Schlssel knnen
  als Prfbytes gelegentlich ausgegeben und beim Einlesen mit dem
  dann aktuellen Schlssel verglichen werden ('putCheck', 'getCheck').
*/

static void CryptInit(void)
/* initializes transcryption */
{
  crypt_R = 55665;
}

static void putPlain(byte plain)
{
  Encrypt(plain);
  putbyte(plain);
}

static int getPlain(bool eof_ok)
{ int plain = getbyte(eof_ok);
  if (plain == EOF) return EOF;
  Encrypt((byte)plain); /* ! */
  return plain;
}

static void putCrypt(byte plain)
{
  putbyte(Encrypt(plain));
}

static int getCrypt(bool eof_ok)
{ int crypt = getbyte(eof_ok);
  if (crypt == EOF) return EOF;
  return Decrypt((byte)crypt);
}

static void putCheck(void)
{
  putbyte((byte)(crypt_R >> 8));
  putbyte((byte)(crypt_R >> 0));
}

static void getCheck(void)
{
  assert1(getbyte(False) == ((crypt_R >> 8) & 0xff),"file '%s' corrupted",fid);
  assert1(getbyte(False) == ((crypt_R >> 0) & 0xff),"file '%s' corrupted",fid);
}

/*D
  Blockung und Validierung

  Um die Leseoperation abzusichern, ist es erforderlich, schon "vor" dem
  Lesen eines Bytes sicher sein zu knnen, dass diese korrekt ist. Im anderen
  Fall bricht ggfls. das Programm vor dem Lesen der Prfbytes bereits
  mit einem anderen Fehler ab, was die eigentliche Ursache verschleiert.

  Der Binrtext wird daher beim Schreiben in "Blcke" eingeteilt. Jeder
  Block wird am Anfang mit der Lnge seines Inhalts und am Ende mit Prfbytes
  versehen. Beim Lesen wird jeweils ein ganzer Block in den Speicher geladen
  bis er auf seine Korrektheit hinsichtlich der Prfbytes untersucht werden
  kann. Dies erfordert die Einrichtung eines Pufferungsmechanismus.

  Die maximale Blocklnge betrgt 'MAXBLKLEN+8' Bytes. Krzere Blcke
  sind mglicht. Der Aufbau eines Blocks sieht im einzelnen so aus:

  |   1 Byte Blockl\ange
  |   N Bytes Text           -- maximal MAXBLKLEN
  |   2 Pr\ufbytes
*/

#define MAXBLKLEN 200
#define BLKEOF    (MAXBLKLEN+1)
static byte BlockTxt[MAXBLKLEN];
static byte BlockLen = 0;
static byte BlockRdr = 0;

static void putBlock(void)
{ int i;
  putCrypt(BlockLen);
  for (i = 0; i < BlockLen; i++)
    putCrypt(BlockTxt[i]);
  putCheck();
  BlockLen = 0;
}

static void getBlock(bool eof_ok)
{ int n, i;
  BlockRdr = 0;
  n = getCrypt(eof_ok);
  if (n == EOF)
  {
    BlockLen = BLKEOF; return;
  }
  BlockLen = (byte)n;
  assert1(BlockLen <= MAXBLKLEN,"file '%s' corrupted",fid);
  for (i = 0; i < BlockLen; i++)
    BlockTxt[i] = getCrypt(False);
  getCheck();
}

static void BlockPutByte(byte x)
{
  if (BlockLen == MAXBLKLEN)
    putBlock();
  BlockTxt[BlockLen++] = x;
}

static int BlockGetByte(bool eof_ok)
{
  if (BlockRdr == BlockLen)
    getBlock(eof_ok);
  if (BlockLen == BLKEOF)
  {
    assert1(eof_ok,"unexpected end of file '%s'",fid);
    return EOF;
  }
  return BlockTxt[BlockRdr++];
}

/* ----------------------------- Kompression ------------------------------ */

/*D
  Datenkompression

  Der Inhalt der Datenseiten wird komprimiert. Ausser dass eine
  Speicherersparnis von ca. 50 Prozent ergibt, ist dies auch angezeigt,
  weil umgekehrt die Verschlsselung der Seiten, den unangenehmen
  Nebeneffekt hat, dass die Bibliothek nachtrglich nicht mehr
  komprimiert werden kann, da sie, statistisch betrachtet, nach
  der Verschlsselung nur noch eine Kette gleichverteilter Bytes ist.
  Sie entlastet im brigen auch bei Binrdateien den Programmieren etwa
  von Platzberlegungen, so kann etwa ein 'long' mit recht verminderter
  Hemmung als 'bool' verwendet werden, wenn dies nur oft genug geschieht.

  Kompressionsverfahren

  Verwendet wird eine Variante der Lempel-Ziv-Welch Kompression.
  Sie besteht aus einer Tabelle, die als Eintrge Binrstrings
  enthlt. Am Anfang enthlt diese Tabelle nur alle Bytes von 0
  bis 255 als Strings der Lnge 1.

  Vom zu komprimierenden Text wird dann der lngste Prfix, der
  in der Tabelle enthalten ist, gesucht. Dessen Nummer (Index)
  wird dann ausgegeben und dieser Prfix, verlngert um das
  unmittelbar folgende Zeichen im Text als neuer String in die
  Tabelle eingetragen. Der Prfix wird dann vom Text entfernt
  und der Rest weiterverarbeitet, bis der Text vollstndig
  komprimiert ist.

  Die Indizies werden als Zahlen mit einer variablen Anzahl an
  Bits ausgegeben. Diese Bitzahl ist die kleinste Zahl, die
  dem aktuellen Fllungsgrad der Tabelle entspricht.
  Begonnen wird also mit 8 Bits, nach dem 1. Eintrag 9. Bit,
  nach dem 256. Eintrag 10. Bit, usw.

  Die Tabelle hat eine feste, hchste Grsse von Eintrgen, die
  neu aufgenommen werden knnen ('MaxEntries'). Falls mehr
  Eintrge aufgenommen werden sollen, wird die Tabelle
  wieder auf ihren Ausgangszustand (nur die einzelnen Bytes)
  zurckgesetzt. Dies ist heuristisch darum angezeigt, da durch
  die Verlngerung der Tabelle auch immer mehr Bits pro Eintrag
  ausgegeben werden mssen, obwohl (ggfls. schon viele Seiten im
  Text spter) inzwischen viele der alten Eintrge gar nicht mehr
  im laufenden Text auftauchen.

  Implementierung

  Das schnelle Auffinden des Prfix ist die Crux des Verfahrens.
  Hierzu wird eine Hash-Tabelle eingesetzt, deren Eintrge Paare
  der Form (Eintrag,Byte) sind. Dies ist nicht nur ein schnelles,
  sondern auch (gemessen an der Grsse) relativ platzsparendes
  Verfahren.

  Dekompression

  Die Dekompression luft im wesentlichen umgekehrt dazu. Wenn ein
  Index gelesen wird, wird dessen String aus der Tabelle gesucht,
  und als dekomprimierter Teil ausgegeben. Ferner bilden das erste
  Zeichen dieses Strings angehngt an den zuletzt vorher ausgegebenen
  String einen neuen Eintrag fr die Tabelle. Der einzige Kniff ist
  der Fall, dass ein Index auflaufen kann, der noch nicht bekannt ist,
  in diesem Fall ist der zuletzt ausgegebene String verlngert um
  sein erstes Zeichen gemeint - wie man sich mit einiger Mhe dann
  doch berlegen kann. Die Tabelle muss natrlich auch bei der
  Dekompression immer wieder auf ihren Anfangswert zurckgesetzt
  werden, wenn sie voll ist. Die Kompression und Dekompression sauber
  zu synchronisieren, war speziell an dieser Stelle am kritischsten.
*/

/* ----------------------------- Kompression ------------------------------ */

#define CharSetSize (1<<8)
#define MaxBitsLen  12
#define HtblSize    (1<<MaxBitsLen)
#define MaxEntries  (HtblSize-CharSetSize)
static struct
{
  short prev;                    /* Vorangehender String ODER! Zeichen       */
  char  c;                       /* Angeh\angtes Zeichen                     */
  short next;                    /* Kette f\ur den Fall eines Hash-Konflikts */
}
       __FAR hentry[MaxEntries];   /* Pr\afix-Tabelle                        */
static short __FAR Htbl[HtblSize]; /* Hash-Tabelle. (-1) ist freier Slot.    */
static short topEntry;             /* Anzahl Eintr\age in Pr\afix-Tabelle    */
static short Log2_top;             /* = ceiling(log2(topEntry+CharSetSize))  */

static short Prev, Leng;                    /* Teilpr\afix und L\ange       */
static byte __FAR res_buffer[MaxEntries];   /* Buffer, f\ur 'unpkByt'       */

static void initEntry(void)
/* ininitalizes 'topEntry' maintaining intergrity of 'Log2_top' */
{
  topEntry = 0; Log2_top = 8;
}

static void incEntry(void)
/* increments 'topEntry' maintaining intergrity of 'Log2_top' */
{
  topEntry += 1;
  if ((topEntry+CharSetSize) > (1 << Log2_top))
    Log2_top += 1;
}

static word HashPair(word prev, char c)
/* hash value needed as 'loc0' for the two following functions */
{
  return (word)((52845 * (prev + c) + 22719) & (HtblSize-1));
}

static short FindPair(word loc0, word prev, char c)
/* finds pair or returns (-1) if not found */
{ short loc;
  for(loc = Htbl[loc0];
      loc != -1 && !(hentry[loc].c == c && hentry[loc].prev == prev);
      loc = hentry[loc].next
     );
  return loc < 0 ? -1 : loc + (short)CharSetSize;
}

static void AddPair(word loc0, word prev, char c)
/* adds pair to table */
{
  assert0(topEntry < MaxEntries,"table overflow");
  hentry[topEntry].c    = c;
  hentry[topEntry].prev = prev;
  hentry[topEntry].next = Htbl[loc0];
  Htbl[loc0]            = topEntry;
  incEntry();
}

static void initTable(void)
/* initalizes the hash table */
{ int i;
  for (i = 0; i < HtblSize; i++)
    Htbl[i] = -1;
  initEntry();
}

/* -- Simulated Interface -- */

static bool someHugePut;
static long BytesIn  = 0;
static long BytesOut = 0;

static void sim_putByte(byte b)
/* */
{
  BlockPutByte(b);
  BytesOut += 1;
}

static int sim_getByte(void)
{
  int res = BlockGetByte(True);
  BytesOut += (res != EOF);
  return res;
}

/* --------------------------------- Bit I/O ------------------------------ */

static unsigned long bit_buffer; /* binary buffer, 4 byte == 32 bit         */
static int bits_buffered;        /* number of bits buffered in 'bit_buffer' */

static void initBits(void)
/* clear bit buffer */
{
  bits_buffered = 0;
}

static void putBits(word bits, int len)
/* put 'len' 'bits' */
{
  bit_buffer = (bit_buffer << len) | bits;
  for (bits_buffered += len; bits_buffered > 7; bits_buffered -= 8)
    sim_putByte((byte)(bit_buffer >> (bits_buffered - 8)));
}

static void flushBits(void)
/* flushes bits remaining in buffer */
{
  if (bits_buffered > 0)
    sim_putByte((byte)(bit_buffer << (8 - bits_buffered)));
}

static int getBits(int len)
/* get 'len' bits */
{ /* This ist tricky! */
  for(bits_buffered -= len; bits_buffered < 0; bits_buffered += 8)
  { int tmp = sim_getByte(); if (tmp == EOF) return -1;
    bit_buffer = (bit_buffer << 8) | tmp;
  }
  return (bit_buffer >> bits_buffered) & ((1<<len)-1);
}

static void packBgn(void)
/* prepare for packing */
{
  BytesIn = 0; BytesOut = 0; initTable(); initBits(); Leng = 0;
}

static void packByt(byte c)
/* pack one byte */
{
  BytesIn += 1;
  if (Leng == 0)                      /* Das erstes Zeichen \uberhaupt ...  */
  {
    Leng = 1; Prev = c;               /* ... beginnt den ersten Pr\afix.    */
  }
  else                                /* Alle weiteren Zeichen:             */
  { word  loc0 = HashPair(Prev,c);    /* Suche String in Tabelle ...        */
    short temp = FindPair(loc0,Prev,c);
    if (temp < 0)                     /* Wenn nicht vorhanden, dann ist ... */
    {                                 /* daf\ur aber der Pr\afix gefunden.  */
      putBits(Prev,Log2_top);         /* Dessen Index ausgeben              */
      if (topEntry < MaxEntries)      /* ... und ...                        */
        AddPair(loc0,Prev,c);         /* (Pr\afix,c) hinzuf\ugen            */
      else                            /* ... oder ...                       */
        initTable();                  /* Tabelle leeren, wenn voll.         */
      Leng = 1; Prev = c;             /* Hier beginnt neuen Pr\afix.        */
    }
    else
    {
      Leng += 1; Prev  = temp;        /* Eintrag bekannt, weiter im Text    */
    }
  }
}

static void packEnd(void)
/* done with packing, flushes all buffered values */
{
  if (Leng > 0)
    putBits(Prev,Log2_top);
  flushBits();
  if (!someHugePut && BytesIn > 1024 && BytesOut > 3*BytesIn/4)
  {
    fprintf(STDERR,"\n");
    fprintf(STDERR,"[Packer Quality Monitor]\n");
    fprintf(STDERR,"!\n");
    fprintf(STDERR,"! Die Packratio betrug weniger als 75%%.\n");
    fprintf(STDERR,"!\n");
    fprintf(STDERR,"! file: %s\n",fid);
    fprintf(STDERR,"! in  : %8ld : 100.00 %%\n",BytesIn);
    fprintf(STDERR,"! out : %8ld : %3ld.%02ld %%\n",BytesOut,
                    100*BytesOut/BytesIn,(10000*BytesOut/BytesIn)%100);
    fprintf(STDERR,"\n");
  }
}

static void unpkBgn(void)
/* prepare for unpacking */
{
  BytesIn = 0; BytesOut = 0; initEntry(); initBits(); Leng = 0;
}

static int unpkByt(void)
/* get one unpacked byte or EOF */
{ int code0,code;
  BytesIn += 1;
  if (Leng > 0)                       /* Obwohl die Dekomprimierung         */
    return res_buffer[--Leng];        /* eigentlich viel einfacher als die  */
  code0 = getBits(Log2_top);          /* Komprimierung ist (es mu\s weder   */
  if (code0 == EOF)                   /* gesucht noch gehasht werden), hat  */
    return EOF;                       /* auch sie so ihre Haken und \Osen.  */
  if (topEntry < MaxEntries)
    hentry[topEntry].prev = code0;
  for(code = code0; code >= CharSetSize; code = hentry[code-CharSetSize].prev)
    res_buffer[Leng++] = hentry[code-CharSetSize].c;
  if (topEntry > 0)
  {
    hentry[topEntry-1].c = code;
    if (code0 == topEntry-1+CharSetSize)
      res_buffer[0] = code;
  }
  if (topEntry == MaxEntries)
    initEntry();
  else
    incEntry();
  return code;
}

static void unpkEnd(void)
/* dummy */
{
}

static void resync(void)
{
  initEntry(); initBits(); Leng = 0;
}

static void flushAllBuffers(void)
{
  if (Leng > 0) putBits(Prev,Log2_top);
  flushBits(); 
  if (BlockLen > 0) putBlock();
  initTable(); initBits(); Leng = 0;
}

/*I----------------------------- Open & Close --------------------------------- */

/*D
  Zustnde

  Um eine korrekte Verwendung des Moduls zu untersttzen, wird ein expliziter
  Zustand gefhrt und vor Ausfhrung jeder Operation berprft. Insgesamt
  gibt es 5 Zustnde

  - geschlossen
  - geffnet zu Lesen oder Schreiben des Dateikopfes
  - bereit zum Lesen oder Schreiben von Daten
*/

#define Closed       (0)
#define Open(MODE)   (1+(MODE))
#define Body(MODE)   (3+(MODE))

#define Get    0
#define Put    1

static int state = Closed;

#define CheckState(STATE) assert1(state == (STATE),"state = %d\n",state)

/*D
  Die folgenden Variablen werden fr 'getReference' und 'putReference'
  bentigt. Sie stehen hier, da sie von den '???Bgn/End'-Routinen
  verwaltet werden.
*/

static MAPTY           puttyp;
static MAP(_,_)        putmap;
static ROW(_)          getmap;

void putBgn(string EnvVar, string FileName, string Ext)
/* open [$'EnvVar'/'FileName''Ext'] to put binary image */
{
  CheckState(Closed);
  fid    = MkPath(EnvVar,FileName,Ext,'e');
  f      = OpnFile(fid,"wb");
  state  = Open(Put);
  srand((int)time(NULL)); /* time dependent ("true") random */
  packBgn();
  puttyp = MAP_newTyp(primCopy,primFree,primEqual,primHash,primCopy,primFree);
  putmap = MAP_newMap(puttyp);
  someHugePut = False;
}

void getBgn(string EnvVar, string FileName, string Ext)
/* open [$'EnvVar'/'FileName''Ext'] to get binary image */
{
  CheckState(Closed);
  fid        = MkPath(EnvVar,FileName,Ext,'p');
  SrcFileLen = Path_FileSize(fid);
  SrcFileGet = 0;
  f          = OpnFile(fid,"rb");
  state      = Open(Get);
  (*IncEvent)((float)0);
  unpkBgn();
  getmap = OT_CREATE_ADT(_);
}

void putEnd(void)
/* completes binary puting */
{
  CheckState(Body(Put));
  packEnd();
  if (BlockLen > 0)
    putBlock();
  fclose(f);
  state = Closed;
  FreeMem(fid);
  MAP_freeMap(putmap);
  MAP_freeTyp(puttyp);
}

void getEnd(void)
/* completes binary geting */
{
  CheckState(Body(Get));
  IncEvent = noIncEvent;
  unpkEnd();
  assert1
  (
    (BlockLen == BLKEOF || BlockLen == BlockRdr) && fgetc(f) == EOF,
    "text after logical end of file '%s'",fid
  );
  fclose(f);
  state = Closed;
  FreeMem(fid);
  OT_DEL_T(getmap);
}

/*I-------------------------------- Header ------------------------------------------ */

/*!!
  There are a lot of reasons to save some informations at the beginning of such a file.

  Beside a short text describing the content of the file, the user want to be sure that
  the file has the expected format. Following an old tradition this will be done by a 
  'Magic'.

  To handle format changes of binary files we introduce a version. The version consists
  of two numbers ('Major', 'Minor'). Binary formats with different major-numbers are
  treated as incompatible. Binary formats with different minor-numbers are treated as
  upward compatible.

  Furthermore this module has an internal version number to track changes of the internal
  format.

  Contrary to the external representation the title will be be saved as null-terminated
  string.

  'getHeader' checks these informations and aborts the operation in the case of an error.
  During the read or write process the current minor-version is accessable via the
  function 'MinorVersion'.
*/

static byte cMinor;   /* current minor version */
static byte cVersion; /* current file  version */

#define Version 2

void putHeader(string Title, string Magic, byte Major, byte Minor)
/* put header */
{ int i;
  CheckState(Open(Put));
  CryptInit();
  while(*Title) putPlain(*Title++); putPlain(0);
  for (i = 0; i < 3; i++) putPlain((byte)rand());
  putCheck();
  BlockLen = 0; BlockRdr = 0;                      /* init blocking */
  while(*Magic ) BlockPutByte(*Magic ++); BlockPutByte(0);
  BlockPutByte(Major);
  BlockPutByte(Minor);
  BlockPutByte(Version);
  putBlock();
  cMinor = Minor;
  state = Body(Put);
}

void getHeaderInfo(string *Com, string *Mag, byte *Ma, byte *Mi, byte *Bv)
/* get header information ( title,magic,major,minor,version ) */
{ int i; byte cc; Sink tmp;
  CheckState(Open(Get));
  CryptInit();
  tmp = Sink_open();
  while( (cc = getPlain(False)), cc != 0)
    Sink_putc(tmp,cc);
  *Com = Sink_close(tmp);
  for (i = 0; i < 3; i++) getPlain(False);
  getCheck();
  BlockLen = 0; BlockRdr = 0;                        /* init blocking */
  tmp = Sink_open();
  while( (cc = BlockGetByte(False)), cc != 0)
    Sink_putc(tmp,cc);
  *Mag = Sink_close(tmp);
  *Ma  = BlockGetByte(False);
  *Mi  = BlockGetByte(False);
  cVersion = BlockGetByte(False);
  *Bv  = cVersion;
  assert1(cVersion == Version,"incompatible binary format of file '%s'",fid);
  assert1(BlockLen == BlockRdr,"file '%s' corrupted",fid);
  state = Body(Get);
}

void getHeader(string Magic, byte Major, byte Minor)
/* validates header */
{ string Tit, Mag; byte Ma, Mi, Bv;
  getHeaderInfo(&Tit, &Mag, &Ma, &Mi, &Bv);
  assert1(strcmp(Magic,Mag)==0,"wrong type of file '%s'",fid);
  assert1(Major == Ma,"incompatible major version of file '%s'",fid);
  assert1(Minor <= Mi,"incompatible minor version of file '%s'",fid);
  cMinor = Mi; FreeMem(Tit); FreeMem(Mag);
}

void getHeaderTitle(string Magic, byte Major, byte Minor, string* Title)
/* validates header, returns title */
{ string Mag; byte Ma, Mi, Bv;
  getHeaderInfo(Title, &Mag, &Ma, &Mi, &Bv);
  assert1(strcmp(Magic,Mag)==0,"wrong type of file '%s'",fid);
  assert1(Major == Ma,"incompatible major version of file '%s'",fid);
  assert1(Minor <= Mi,"incompatible minor version of file '%s'",fid);
  cMinor = Mi; FreeMem(Mag);
}

short MinorVersion(void)
/* 'Minor' of the file */
{
  return cMinor;
}

/*I============================= Data types =============================== */

/*!!
  Actually the following data types are supported.

  | TYPE      | CTYPE               |
  +-----------+---------------------+----------------------------
  | Byte      | unsigned char       |
  | Word      | unsigned short int  |   Intrinsic C-data types
  | Long      | signed long int     |
  | Int       | signed int          |
  +-----------+---------------------+----------------------------
  | String    | (char *)            |   Strings
  | Symbol    | symbol              |   Symbols
  | Function  | (? (*)())           |   Functions
  | Abstract  | (?)                 |   "Objects"
  | StdCPtr   | (?*)                |   References
*/

/*I----------------------------- Plain values ----------------------------- */

void putByte(byte v)
/* put 'v' to file */
{
  CheckState(Body(Put));
  packByt(v);
}

void getByte(byte *v)
/* get 'v' from file */
{ int val;
  CheckState(Body(Get));
  val = unpkByt();
  assert1(val != EOF,"unexpected end of file '%s'",fid);
  *v = (byte)val;
}

int getByte_or_EOF(void)
/* get byte or EOF from file */
{ CheckState(Body(Get)); return unpkByt();
}

void putWord(short v)
/* put 'v' to file; msb first */
{
  putByte((byte)(v >> 8));
  putByte((byte)(v >> 0));
}

void getWord(short *v)
/* get 'v' from file; msb first */
{ byte msb,lsb;
  getByte(&msb);
  getByte(&lsb);
  *v = ((short)msb << 8) | lsb;
}

void putLong(long v)
/* put 'v' to file; msw first */
{
  putWord((short)(v >> 16));
  putWord((short)(v >>  0));
}

void getLong(long *v)
/* get 'v' from file; msw first */
{ word msw,lsw;
  getWord((short *)&msw);
  getWord((short *)&lsw);
  *v = ((long)msw << 16) | lsw;
}

void putULong(unsigned long v)
/* put 'v' to file; msw first */
{ 
  putWord((short)(v >> 16));
  putWord((short)(v >>  0));
}

void getULong(unsigned long *v)
/* get 'v' from file; msw first */
{ word msw,lsw;
  getWord((short *)&msw);
  getWord((short *)&lsw);
  *v = ((unsigned long)msw << 16) | lsw;
}

void putInt(int v)
/* put 'v' to file */
{
  assert1( ( ((long) v) & 0xffff0000L ) == 0          ||
           ( ((long) v) & 0xffff0000L ) == 0xffff0000L,
          "'%d' erfordert mehr als 2 Byte zur Darstellung (nimm 'putLong')",v);
  putWord((short)v);
}

void getInt(int* v)
/* get 'v' from file */
{ short v0;
  getWord(&v0);
  *v = v0;
}

void putString(string v)
/* put 'v' to file; length byte first */
{
  putInt(strlen(v));
  for(; *v; v++)
    putByte(*v);
}

void getString(string *v)
/* get 'v' from file; length first; allocs memory */
{ short len,i;
  getWord(&len);
  *v = (string) NewMem((len+1) * sizeof(char));
  for(i = 0; i < len; i++)
    getByte((byte*)&(*v)[i]);
  (*v)[len] = '\0';
}

/*I----------------------------- Huge binaries ------------------------------- */

/*!
  To save and load large binary data blocks the following functions can be used.
*/

#define HugeBufferLen ((size_t)(1L<<15))

void putHuge(HugeCPtr v, long len)
/* put 'len' bytes to file */
{ byte *       buf = (byte*)NewMem(HugeBufferLen);
  byte __HUGE* img = (byte __HUGE*)v;
  long         iln;
  putLong(len);
  flushAllBuffers(); /* flush buffered data */
  BytesIn  += len;
  BytesOut += len;
  someHugePut = True;
  for(iln = len; iln >= HugeBufferLen; iln -= HugeBufferLen)
  { size_t i;
    for (i = 0; i < HugeBufferLen; i++)
      buf[i] = Encrypt(*img++);
    fwrite(buf,HugeBufferLen,1,f);
  }
  if (iln > 0)
  { size_t i;
    for (i = 0; i < iln; i++)
      buf[i] = Encrypt(*img++);
    fwrite(buf,(size_t)iln,1,f);
  }
  FreeMem(buf);
  putCheck();
}

void getHuge(HugeCPtr *v, long *len)
/* get 'len' bytes from file */
{ byte *       buf = (byte*)NewMem(HugeBufferLen);
  byte __HUGE* img;
  long         iln;
  getLong(len);
  resync();
  BytesIn  += *len;
  BytesOut += *len;
  *v = NewHMem(*len);
  img = (byte __HUGE*)*v;
  for(iln = *len; iln >= HugeBufferLen; iln -= HugeBufferLen)
  { size_t i;
    fread(buf,HugeBufferLen,1,f);
    for (i = 0; i < HugeBufferLen; i++)
      (*img++) = Decrypt(buf[i]);
  }
  if (iln > 0)
  { size_t i;
    fread(buf,(size_t)iln,1,f);
    for (i = 0; i < iln; i++)
      (*img++) = Decrypt(buf[i]);
  }
  FreeMem(buf);
  getCheck();
}

/*I-------------------------------- Symbols ------------------------------- */

/*!
  Symbols are externally represented as strings.
*/

void putSymbol(symbol v)
/* put a symbol to file */
{
  putString(symbolToString(v));
}

void getSymbol(symbol *v)
/* get a symbol from file */
{ string s;
  getString(&s);
  *v = stringToSymbol(s);
  FreeMem(s);
}

/*I------------------------------- Functions ------------------------------ */

/*!
   For technical reasons the functions must be defined in a global table.<br>
   They are externally represented by a symbolic name representing the key
   to the function table entry. ( see also <a href="glo_tab.htm">[glo_tab]</a> )
*/

void putFunction(StdCPtr v)
/* put a function to file
   raises error if 'v' not 'Glo'bally defined 
*/
{
  putSymbol(Glo_key(v));
}

void getFunction(StdCPtr *v)
/* get a function from file
   raises error if 's' not 'Glo'bally defined 
*/
{ symbol s;
  getSymbol(&s);
  *v = Glo_val(StdCPtr,s);
}

/*I---------------------------- Abstract types ---------------------------- */

/*!!
   In the case of a generic data type ( e.g. 'List(Alpha)') a 'put'-function
   typically looks like:

   |  void putList(List(Alpha) v, void putAlpha(Alpha v))
   |  {
   |    putInt(List_length(v));
   |    for (; !List_null(v); v = List_rest(v))
   |      putAlpha(List_first(Alpha,v));
   |  }

   In the case of a heterogen parameter type ("Object") the user has to save 
   the corresponding 'get'-function together with the value.
*/

void putAbstract(Abs_T v, void putData(Abs_T v), void getData(Abs_T *v))
/* put abstract data to file */
{
  putFunction((StdCPtr)getData);
  putData(v);
}

void getAbstract(Abs_T *v)
/* get abstract data from file */
{ StdCPtr getData;
  getFunction(&getData);
  ((void (*)(Abs_T *))getData)(v);
}

/*I-------------------------------- Pointer ------------------------------------------- */

/*!!
   References to multiple or cyclic referenced structures ( except symbols and functions )
   can't be simply expanded if the representation should be unique.
   For cases like this we support the following function.

   |  void putReference(Abs_T  v, void putData(Abs_T v));

   This function outputs a reference number for this structur and only in the case of the 
   first reference the structure values.
*/

void putReference(Abs_T v, void putData(Abs_T v))
/* put a pointer to file */
{
  if (MAP_defined(putmap,v))
  {
    putLong(MAP_apply(long,putmap,v));
  }
  else
  { long id = MAP_count(putmap);
    MAP_dfndom(putmap,v,id);
    putLong(id);
    putData(v);
  }
}

void getReference(Abs_T *v, void getData(Abs_T *v))
/* get a pointer from file */
{ long getid;
  getLong(&getid);
  if (getid < OT_CNT(getmap))
  {
    *v = OT_GET(Abs_T,getmap,getid);
  }
  else
  {
    assert0(getid == OT_CNT(getmap),"out of sync");
    OT_PUSH(NULL,getmap);
    getData(v);
    OT_UPD(Abs_T,getmap,getid,*v);
  }
}
