/* ------------------------------------------------------------------------ */
/*                                                                          */
/* [ptm_pp.c]          Parse Tree Transformation & Pretty Printing          */
/*                                                                          */
/* Copyright (c) 1995 by D\olle, Manns                                      */
/* ------------------------------------------------------------------------ */

/*FIXME: layout, unicode support for pretty printer */

#include "ptm.h"

#include "cfg_dfn.h"
#include "prs_dfn.h"
#include "prs.h"
#include "hmap.h"
#include "hset.h"
#include "pathes.h"

/*!!
   The module [ptm_pp] supports some means for source-source-transformations.

   1) It performs a tree-to-tree transformation based on the corresponding 
      grammar specifications. It handles comments, too.

      source tree based on CFG 1
      specification of CFG 2
      -------------------------------------------->  concrete target tree
      abstraction(CFG 1)   = abstraction(CFG 2)      based on CFG 2
      regexp(token(CFG 1)) = regexp(token(CFG 2))

   2) It provides the pretty printing of parse trees. This is usefull after
      a transformation, because the target tree doesn't contain any position
      information.

   Note:
   This module is still under construction. The main things todo refer to the
   layout results which are some times broken and a mechanism for literal
   conversion.
   The interface functions are not reentrant. You cannot perform multiple
   transformations at a time.
*/

/*I---------------------------- Macros ------------------------------------- */

/*! 
  <p><b>Token separation types</b>
*/

/*DOC_INTERFACE*/
#define PTP_SEP_NON 0 // none
#define PTP_SEP_SPC 1 // space
#define PTP_SEP_IND 2 // indent
#define PTP_SEP_ROW 3 // row

#define PTP_MAX_COL 78
#define PTP_TAB_LEN 1

#define ROWS(l)     ( (short)(((unsigned long)(l)) >> 16) )
#define COLS(l)     ( (short)(((unsigned long)(l)) & 0xFFFF) )

static symbol  language, outfile;
static int     linlen    = PTP_MAX_COL;

static bool    nlflg     = False;
static bool    tosync    = True;
static PT_Itr  lookahead = (PT_Itr)NULL;
static PT_Term currtok   = (PT_Term)NULL;

static MAPTY mtyp0, mtyp1, mtyp2;

/* NT |-> PRD |-> PRDSYM |-> SYMFRM */
static MAP(symbol,MAP(symbol,ROW(int))) frmmap; 

/* NT |-> PRD |-> PRDSYMBOLS */
static MAP(symbol,ROW(ROW(symbol))) ntmmap; 

/* TK |-> { TK } -- Token, which need not be seperated */
static MAP(symbol,MAP(symbol,int)) tkmmap;

static MAP(symbol,_) visited;

/*I---------------------------- Init & Quit -------------------------------- */

void PTP_init(PLR_Tab tab)
/* initializes transformation and pretty printing 
   based on parse table 'tab' as target grammar specification
*/
{ int pcnt = PLR_prodCnt(tab), scnt, i, j, n, f;
  outfile  = stringToSymbol("");
  language = stringToSymbol(PLR_language(tab));
  mtyp0 = MAP_newTyp(primCopy,primFree,primEqual,primHash,primCopy,primFree);
  mtyp1 = MAP_newTyp(primCopy,primFree,primEqual,primHash,primCopy,HMP_freeMap);
  mtyp2 = MAP_newTyp(primCopy,primFree,primEqual,primHash,primCopy,OT_delT);
  visited = MAP_newMap(mtyp0);
  frmmap  = MAP_newMap(mtyp1);
  ntmmap  = MAP_newMap(mtyp2);
  tkmmap  = MAP_newMap(mtyp1);
  for( i=0; i < pcnt; ++i )
  { symbol s1, s2; 
    MAP(symbol,ROW(int)) m; ROW(_) t, t2, t3;
    n    = PLR_prodNonTerm(tab,i);
    scnt = PLR_prodSymCnt(tab,i);
    s1 = stringToSymbol(PLR_symName(tab,n));
    s2 = stringToSymbol(PLR_prodName(tab,i));
    if( ! MAP_defined(frmmap,s1) )
    {
      m = MAP_newMap(mtyp2);
      t = OT_CREATE(ROW(ROW(symbol)),primCopy,OT_delT,primEqual);
      MAP_dfndom(frmmap,s1,m);
      MAP_dfndom(ntmmap,s1,t);
    }
    else 
    {
      m = MAP_apply(MAP(symbol,ROW(int)),frmmap,s1);
      t = MAP_apply(ROW(ROW(symbol)),ntmmap,s1);
    }
    t2 = OT_CREATE_ADT(int);
    t3 = OT_CREATE_ADT(symbol);
    MAP_dfndom(m,s2,t2);
    OT_T_INS(t2,PLR_prodMethod(tab,i));
    OT_T_INS(t,t3);
    OT_T_INS(t3,s2);
    for( j=0; j < scnt; ++j )
    { int ps = PLR_prodSymbol(tab,i,j);
      OT_T_INS(t3,stringToSymbol(PLR_symName(tab,ps)));
      OT_T_INS(t3,PLR_symType(tab,ps));
      f = PLR_prodSymFrm(tab,i,j); OT_T_INS(t2,f);
    }
  }
}

void PTP_quit(void)
/* terminates transformation and pretty printing */
{
  MAP_freeMap(visited);
  MAP_freeMap(frmmap);
  MAP_freeMap(ntmmap);
  MAP_freeMap(tkmmap);
  MAP_freeTyp(mtyp2);
  MAP_freeTyp(mtyp1);
  MAP_freeTyp(mtyp0);
}

void PTP_set_line(int len)
/* set 'len' as line length for pretty printing */
{
  assert0((linlen = len) > 0,"length <= 0");
}

void PTP_tok_sep(symbol t_lft, symbol t_rgt, int sep)
/* specifies the kind of seperation ( 'sep' )
   between the token 't_rgt' and 't_lft' 
*/
{ MAP(symbol,int) map;
  assert(sep >= PTP_SEP_NON && sep <= PTP_SEP_ROW,"invalid seperator");
  if( ! MAP_defined(tkmmap,t_lft) )
  {
    map = MAP_newMap(mtyp0);
    MAP_dfndom(tkmmap,t_lft,map);
  }
  else map = MAP_apply(MAP(symbol,map),tkmmap,t_lft);
  if( MAP_defined(map,t_rgt) ) MAP_upddom(map,t_rgt,sep);
  else MAP_dfndom(map,t_rgt,sep);
}

/*I------------------------- Term Pretty Printer --------------------------- */

/* ------------------------- Pass 1: Symbol Positions ---------------------- */

static void PTP_swap(StdCPtr* a, StdCPtr* b)
{ StdCPtr h = *a;
  *a = *b; *b = h;
}

static unsigned long PTP_toklen(short row, short col)
{ unsigned long l, r, c;
  r = ((unsigned long)row) << 16;
  r = r & 0xFFFF0000;
  c = ((unsigned long)col) & 0x0000FFFF;
  l = r | c;
  return l;
}

static unsigned long PTP_vallen(string v)
{ short tr = 0, tc = 0;
  unsigned int l = strlen(v), i;
  for( i=0; i < l; ++i )
  {
    if( v[i] == '\n' ) 
    { 
      ++tr; tc = 0; 
    }
    else 
    if( v[i] == '\t' ) { tc += PTP_TAB_LEN; }
    else ++tc;
  }
  return PTP_toklen(tr,tc);
}

static unsigned long PTP_add_len
                     (
                       unsigned long l1, unsigned long ls, unsigned long l2
                     )
{ short r, c;
  r = ROWS(l1) + ROWS(ls) + ROWS(l2);
  c = COLS(l1) + COLS(ls) + COLS(l2);
  if( c >= linlen )
  {
    ++r; c = COLS(l2);
  }
/*XXX
  else c += ( COLS(l1) > 0 && COLS(l2) > 0 );
*/
  return PTP_toklen(r,c);
}

static void PTP_setNextTok(PT_Itr itr)
{ int state = PT_stateIT(itr);
  if( state == PT_ROOT || state == PT_PRAE || state == PT_POST )
  {
    while( PT_df_getIT(itr) )
    { 
      if( PT_stateIT(itr) == PT_PRAE && ! PT_isNonTerm(PT_termIT(itr)) )
        break;
    }
  }
}

static PT_Term PTP_getNextTok(PT_Itr itr)
{ int state = PT_stateIT(itr);
  return ( state == PT_ROOT || state == PT_PRAE || state == PT_POST )
         ? PT_termIT(itr) : (PT_Term)NULL;
}

static short PTP_toksep(PT_Term l, PT_Term r)
{ short res = PTP_SEP_NON;
  if( l != (PT_Term)NULL  && r != (PT_Term)NULL ) 
  { 
    res = PT_isComment(l) ? PTP_SEP_IND : PTP_SEP_SPC;
    if( MAP_defined(tkmmap,PT_symbol(l)) )
    { MAP(symbol,int) map = 
        MAP_apply(MAP(symbol,int),tkmmap,PT_symbol(l));
      if( MAP_defined(map,PT_symbol(r)) )
        res = MAP_apply(int,map,PT_symbol(r));
    }
  }
  return res;
}

static short PTP_colsep(short row, short col)
{ PT_Term nexttok = PTP_getNextTok(lookahead);
  short sep = PTP_toksep(currtok,nexttok); 
  return (short)( 
           sep >= PTP_SEP_SPC     && 
           row == PT_row(currtok) && 
           col <  PT_col(currtok) + 
                  COLS((unsigned long)PT_file(currtok)) + sep
         );
}

static void PTP_tokpos(PT_Term t, short* row, short* col, short ind)
{ string v = symbolToString(PT_value(t));
  unsigned long tl = PTP_vallen(v);
  short r = *row, c = *col; 
  if( COLS(tl) + c >= linlen )
  {
    r += ( ! nlflg ); 
    c = (( COLS(tl) + ind >= linlen ) ? 0 : ind);
  }
  PT_updPos(t,(symbol)tl,r,c);
  *row = r + ROWS(tl); *col = c + COLS(tl);
}

static void PTP_updpos(PT_Term tree, short row, short* col)
{ 
  if( tree != (PT_Term)NULL )
  { 
    if( tosync )
    {
      *col = *col + PTP_colsep(row,*col);
      tosync = False;
    }
    PT_updPos(tree,PT_file(tree),row,*col);
    if( PT_isNonTerm(tree) )
    {
      if( ! PT_isXaron(tree) )
      { PT_Term ts; 
        for( ts = PT_parts(tree); ts != (PT_Term)NULL; ts = PT_next(ts) )
          PTP_updpos(ts,row,col);
      }
      else
      { List(PT_Term) ts;
        for( ts = XPT_parts(tree); ! empty(ts); ts = rst(ts))
          PTP_updpos(fst(PT_Term,ts),row,col);
      }
    }
    else 
    {
      currtok = tree;
      PTP_setNextTok(lookahead);
      *col = *col + COLS((unsigned long)PT_file(tree));
      tosync = True;
    }
  }
}

static unsigned long PTP_cfgsep(PT_Term t, ROW(int) ps, long* idx)
{ short row = 0, col = 0;
  if( ! PT_isComment(t) && ps != (ROW(int))NULL && OT_CNT(ps) > 1 )
  { 
    if( *idx > 1 ) 
    { int f; 
      f = OT_GET(int,ps,*idx);
      if( f <= 0 ) 
      {
        row = 1; 
        col = ABS(f);
      }
    }
    *idx = *idx + 1;
  }
  return PTP_toklen(row,col);
}

static void PTP_sync_to_cfg
            (
              unsigned long sl, short* row, short* col, short* ind
            )
{
  if( ROWS(sl) )
  {
    if( ! nlflg )
    {
      *row  = *row + 1; 
      nlflg = True;
    }
    *col = *ind = *ind + COLS(sl);
  }
}

static unsigned long PTP_sync_to_tok
                     (
                       unsigned long sl, short sep,
                       short* row, short* col, short* ind
                     )
{ /* res = seperation between current and next token */
  unsigned long res = 0L; 
  if( sep == PTP_SEP_IND || sep == PTP_SEP_ROW )
  { short r = 1 + ( sep == PTP_SEP_ROW );
    *row  = *row + r; 
    if( ROWS(sl) > 0 )
    {
      *col = *ind = *ind + COLS(sl);
    }
    else *col = *ind;
    nlflg = True;
    res = PTP_toklen(r,0); 
  }
  else
  {
    if( ROWS(sl) > 0 )
    {
      *row = *row + 1; 
      *col = *ind = *ind + COLS(sl);
      nlflg = True;
    }
    else 
    { short c = ( sep == PTP_SEP_SPC );
      *col = *col + c;
      res = PTP_toklen(0,c); 
    }
  }
  return res;
}

static unsigned long PTP_pass1_aux
                     (
                       PT_Term t, ROW(int) ps, long* idx, 
                       short* row, short* col, short* ind
                     )
{ unsigned long sl = PTP_cfgsep(t,ps,idx), res = 0L;
  PT_Term nexttok = PTP_getNextTok(lookahead);
  short sep = PTP_toksep(currtok,nexttok); 
  if( tosync )
  {
    res = PTP_sync_to_tok(sl,sep,row,col,ind);
    tosync = False;
  }
  else PTP_sync_to_cfg(sl,row,col,ind);
  if( ! PT_isNonTerm(t) )
  {
    assert0(PTP_colsep(*row,*col) == 0,"missing token seperator");
    PTP_tokpos(t,row,col,*ind);
    currtok = t;
    PTP_setNextTok(lookahead);
    tosync = True;
    nlflg = False;
  }
  return res;
}

static void PTP_pass1(PT_Term tree, short* row, short* col, short ind)
{ 
  if( tree != (PT_Term)NULL )
  {
    if( PT_isNonTerm(tree) )
    { symbol p = PT_product(tree), n = PT_symbol(tree); 
      short r = *row, c = *col, pi; long i = 1L; 
      unsigned long tl = 0L, tl_aux; bool nl = nlflg; 
      ROW(int) ps = (ROW(int))NULL;
      PT_Itr it = PT_copyIT(lookahead);
      PT_Term ct = currtok;
      if( MAP_defined(frmmap,n) )
      { MAP(symbol,ROW(int)) m = MAP_apply(MAP(symbol,ROW(int)),frmmap,n);
        ps = MAP_apply(ROW(int),m,p);
        if( OT_GET(int,ps,0) == PLR_PRD_GRP ) ++r;
      }
      if( ! PT_isXaron(tree) )
      { PT_Term ts;
        for( ts = PT_parts(tree); ts != (PT_Term)NULL; ts = PT_next(ts) )
        {
          pi = *col;
          tl_aux = PTP_pass1_aux(ts,ps,&i,&r,&c,&pi);
          PTP_pass1(ts,&r,&c,pi);
          tl = PTP_add_len(tl,tl_aux,(unsigned long)PT_file(ts));
        }
        if( ps != (ROW(int))NULL            && 
            OT_GET(int,ps,0) == PLR_PRD_REC && 
            ROWS(tl) == 0                   &&
            COLS(tl) + *col < linlen         )
        { 
          c = *col; r = *row;
          PTP_swap((StdCPtr*)(&lookahead),(StdCPtr*)(&it)); 
          PTP_swap((StdCPtr*)(&currtok),(StdCPtr*)(&ct)); 
          PTP_swap((StdCPtr*)(&nlflg),(StdCPtr*)(&nl));
          for( ts = PT_parts(tree); ts != (PT_Term)NULL; ts = PT_next(ts) )
            PTP_updpos(ts,r,&c);
          PTP_swap((StdCPtr*)(&it),(StdCPtr*)(&lookahead)); 
          PTP_swap((StdCPtr*)(&ct),(StdCPtr*)(&currtok)); 
          nlflg = False;
        }
      }
      else
      { List(PT_Term) ts; PT_Term t;
        for( ts = XPT_parts(tree); ! empty(ts); ts = rst(ts))
        { 
          pi = *col;
          t = fst(PT_Term,ts);
          tl_aux = PTP_pass1_aux(t,ps,&i,&r,&c,&pi);
          PTP_pass1(t,&r,&c,pi);
          tl = PTP_add_len(tl,tl_aux,(unsigned long)PT_file(t));
        }
        if( ps != (ROW(int))NULL            && 
            OT_GET(int,ps,0) == PLR_PRD_REC && 
            ROWS(tl) == 0                   &&
            COLS(tl) + *col < linlen         )
        {
          c = *col; r = *row;
          PTP_swap((StdCPtr*)(&lookahead),(StdCPtr*)(&it)); 
          PTP_swap((StdCPtr*)(&currtok),(StdCPtr*)(&ct)); 
          PTP_swap((StdCPtr*)(&nlflg),(StdCPtr*)(&nl));
          for( ts = XPT_parts(tree); ! empty(ts); ts = rst(ts))
            PTP_updpos(fst(PT_Term,ts),r,&c);
          PTP_swap((StdCPtr*)(&it),(StdCPtr*)(&lookahead)); 
          PTP_swap((StdCPtr*)(&ct),(StdCPtr*)(&currtok)); 
          nlflg = False;
        }
      }
      PT_updPos(tree,(symbol)tl,*row,*col);
      *row = r; *col = c;
      PT_delIT(it);
    }
  }
}

/* ------------------------- Pass 2: Pretty Printing ----------------------- */

static void PTP_pass3(PT_Term tree, FILE* fp, short* row, short* col)
{
  if( tree != (PT_Term)NULL )
  {
    PT_updPos(tree,outfile,PT_row(tree),PT_col(tree));
    if( PT_isNonTerm(tree) )
    {
      if( ! PT_isXaron(tree) )
      { PT_Term ts;
        for( ts = PT_parts(tree); ts != (PT_Term)NULL; ts = PT_next(ts) )
        {
          PTP_pass3(ts,fp,row,col);
        }
      }
      else
      { List(PT_Term) ts; PT_Term t;
        for( ts = XPT_parts(tree); ! empty(ts); ts = rst(ts))
        { 
          t = fst(PT_Term,ts);
          PTP_pass3(t,fp,row,col);
        }
      }
    }
    else
    { string v = symbolToString(PT_value(tree));
      unsigned int l = strlen(v), i;
      short r  = PT_row(tree), 
            c  = PT_col(tree), 
            tr = *row, 
            tc = ( tr < r ) ? 0 : *col; 
      assert
      ( 
        r >= tr && c >= tc,
        "wrong positions: %d >= %d, %d >= %d",r,tr,c,tc
      );
      while( tr < r ) { fprintf(fp,"\n"); ++tr; }
      while( tc < c ) { fprintf(fp," ");  ++tc; }
      tr = tc = 0;
      for( i=0; i < l; ++i )
      {
        if( v[i] == '\n' ) 
        { fprintf(fp,"\n"); ++tr; tc = 0; }
        else 
        if( v[i] == '\t' ) 
        { fprintf(fp,"    "); tc += PTP_TAB_LEN; }
        else 
        { fprintf(fp,"%c",v[i]); ++tc; }
      }
      *row = r + tr; *col = c + tc;
    }
  }
}

void PTP_pp(PT_Term tree, FILE* fp)
/* performs pretty printing on parse term 'tree' 
   ( output to file 'fp' )
*/
{ short row = 0, col = 0;
  lookahead = PT_newIT(tree);
  PTP_setNextTok(lookahead);
  currtok   = (PT_Term)NULL;
  nlflg     = False;
  tosync    = True;
  PTP_pass1(tree,&row,&col,0);
  PT_delIT(lookahead);
/*XXX
fprintf(STDOUT,"LINELEN %d\n",linlen);
PT_prind(tree,0);
fprintf(STDOUT,"\n");
*/
  row = col = 0;
  PTP_pass3(tree,fp,&row,&col);
}

/*I------------------------- Tree Transformation --------------------------- */

static PT_Term PTP_nt_transform(PT_Term tree, symbol nt); /*forward*/

static List(PT_Term) XPTP_parts_transform
                     (
                       List(PT_Term) parts, 
                       ROW(symbol) psrow, long idx, bool* ok
                     )
/* abstract src parts, dst members |--> concrete dst parts ( Xaron-Typ ) */
{ List(PT_Term) res; PT_Term t = (PT_Term)NULL; 
  long cnt = OT_CNT(psrow), i;
  symbol psym; short ptyp;
  if( cnt > idx + 1 )
  { 
    psym = OT_GET(symbol,psrow,idx);
    ptyp = OT_GET_short(psrow,idx+1);
    if( CFG_KEY(ptyp) )
    {
      res = XPTP_parts_transform(parts,psrow,idx+2,ok);
      return (*ok)?XPT_consT(XPT_newTNode(psym,ptyp,psym),res):res;
    }
  }
  if( ! empty(parts) )
  { 
    t = fst(PT_Term,parts);
    if( PT_isKey(t) )
      return XPTP_parts_transform(rst(parts),psrow,idx,ok);
    else
    if( PT_isComment(t) )
    {
      res = XPTP_parts_transform(rst(parts),psrow,idx,ok);
      return (*ok)?XPT_consT(PT_copyNode(t),res):res;
    }
  }
  if( cnt > idx + 1 )
  { 
    psym = OT_GET(symbol,psrow,idx);
    ptyp = OT_GET_short(psrow,idx+1);
    if( CFG_TOK(ptyp) )
    {
      if( empty(parts) || ptyp != PT_symtyp(t) || PT_symbol(t) != psym )
      {
        *ok = False; return (List(PT_Term))NULL;
      }
      res = XPTP_parts_transform(rst(parts),psrow,idx+2,ok);
      return (*ok)?XPT_consT(PT_copyNode(t),res):res;
    }
    else
    {
      assert0( CFG_NTM(ptyp), "" );
      if( empty(parts) || ptyp != PT_symtyp(t) )
      {
        *ok = False; return (List(PT_Term))NULL;
      }
      else
      { ROW(PT_Term) coms = OT_CREATE_ADT(PT_Term);
        t = PT_ignkey_Skip(t,coms);
        t = PTP_nt_transform(t,psym);
        if( t == (PT_Term)NULL )
        {
          *ok = False; OT_DEL_T(coms); return (List(PT_Term))NULL;
        }
        res = XPTP_parts_transform(rst(parts),psrow,idx+2,ok);
        if( ! (*ok) )
        {
          PT_delT(t); OT_DEL_T(coms); return res;
        }
        res = XPT_consT(t,res);
        for( i=OT_CNT(coms)-1; i >= 0; --i )
          res = XPT_consT(PT_copyNode(OT_GET(PT_Term,coms,i)),res);
        OT_DEL_T(coms);
        return res;
      }
    }
  }
  else
  {
    if( ! empty(parts) ) *ok = False;
    return (List(PT_Term))NULL;
  }
}

static PT_Term PTP_parts_transform
               (
                 PT_Term parts, ROW(symbol) psrow, long idx, bool* ok
               )
/* abstract src parts, dst members |--> concrete dst parts */
{ PT_Term t, res; 
  long cnt = OT_CNT(psrow), i;
  symbol psym; short ptyp;
  if( cnt > idx + 1 )
  { 
    psym = OT_GET(symbol,psrow,idx);
    ptyp = OT_GET_short(psrow,idx+1);
    if( CFG_KEY(ptyp) )
    {
      res = PTP_parts_transform(parts,psrow,idx+2,ok);
      return (*ok)?PT_consT(PT_newTNode(psym,ptyp,psym),res):res;
    }
  }
  if( parts != (PT_Term)NULL )
  { 
    if( PT_isKey(parts) )
      return PTP_parts_transform(PT_next(parts),psrow,idx,ok);
    else
    if( PT_isComment(parts) )
    {
      res = PTP_parts_transform(PT_next(parts),psrow,idx,ok);
      return (*ok)?PT_consT(PT_copyNode(parts),res):res;
    }
  }
  if( cnt > idx + 1 )
  { 
    psym = OT_GET(symbol,psrow,idx);
    ptyp = OT_GET_short(psrow,idx+1);
    if( CFG_TOK(ptyp) )
    {
      if( parts == (PT_Term)NULL   || 
          ptyp != PT_symtyp(parts) || 
          PT_symbol(parts) != psym  )
      {
        *ok = False; return (PT_Term)NULL;
      }
      res = PTP_parts_transform(PT_next(parts),psrow,idx+2,ok);
      return (*ok)?PT_consT(PT_copyNode(parts),res):res;
    }
    else
    {
      assert0( CFG_NTM(ptyp), "" );
      if( parts == (PT_Term)NULL || ptyp != PT_symtyp(parts) )
      {
        *ok = False; return (PT_Term)NULL;
      }
      else
      { ROW(PT_Term) coms = OT_CREATE_ADT(PT_Term);
        t = PT_ignkey_Skip(parts,coms);
        t = PTP_nt_transform(t,psym);
        if( t == (PT_Term)NULL )
        {
          *ok = False; OT_DEL_T(coms); return (PT_Term)NULL;
        }
        res = PTP_parts_transform(PT_next(parts),psrow,idx+2,ok);
        if( ! (*ok) )
        {
          PT_delT(t); OT_DEL_T(coms); return res;
        }
        res = PT_consT(t,res);
        for( i=OT_CNT(coms)-1; i >= 0; --i )
          res = PT_consT(PT_copyNode(OT_GET(PT_Term,coms,i)),res);
        OT_DEL_T(coms);
        return res;
      }
    }
  }
  else
  {
    if( parts != (PT_Term)NULL ) *ok = False;
    return (PT_Term)NULL;
  }
}

static void PTP_move_visited(MAP(symbol,_) dst, MAP(symbol,_) src)
{ HMP_Dom dom; HMP_Itr itr;
  MAP_rmvall(dst);
  MAP_forItr(dom,itr,src) MAP_dfndom(dst,dom,_);
  MAP_rmvall(src);
}

/* for diagnosis, see 'PTP_nt_transform' and 'PTP_transform' */
static int    last_lin;
static int    last_col;
static symbol last_nt;
static symbol last_prod;

static PT_Term PTP_nt_transform(PT_Term tree, symbol nt)
/* abstract src tree, dst nt |--> concrete dst tree */
{ 
  if (last_lin < PT_row(tree) || 
      (last_lin == PT_row(tree) && last_col < PT_col(tree)))
  {
    last_lin  = PT_row(tree);
    last_col  = PT_col(tree);
    last_nt   = nt;
    last_prod = PT_product(tree);
  }
  if( tree != (PT_Term)NULL )
  { ROW(ROW(symbol)) prow; symbol a_prod, c_prod; 
    long cnt, cnt2, i, j; bool xflg = PT_isXaron(tree);
    assert0( PT_isNonTerm(tree) && MAP_defined(ntmmap,nt), "" );
    a_prod = PT_product(tree); 
    prow   = MAP_apply(ROW(ROW(symbol)),ntmmap,nt);
    cnt    = OT_CNT(prow);
    for( i=0; i < cnt; ++i )
    { ROW(symbol) psrow = OT_GET(ROW(symbol),prow,i); 
      StdCPtr     parts = (StdCPtr)NULL; 
      assert0( ( cnt2 = OT_CNT(psrow) ) > 0, "" );
      c_prod = OT_GET(symbol,psrow,0);
      if( a_prod == c_prod )
      { MAP(symbol,_) l_visited = MAP_newMap(mtyp0);
        bool ok = True;
        PTP_move_visited(l_visited,visited);
        parts = xflg ? XPTP_parts_transform(XPT_parts(tree),psrow,1,&ok)
                     : PTP_parts_transform(PT_parts(tree),psrow,1,&ok);
        if( ok ) 
        {
          MAP_rmvall(visited); MAP_freeMap(l_visited);
          return xflg ? XPT_newNtNode(nt,c_prod,(List(PT_Term))parts)
                      : PT_newNtNode(nt,c_prod,(PT_Term)parts);
        }
        else
        {
          PTP_move_visited(visited,l_visited); MAP_freeMap(l_visited);
        }
      }
      else
      if( ignProd(symbolToString(c_prod)) ) 
      { 
        for( j=cnt2-2; j >= 1; j -= 2 ) 
        { symbol psym = OT_GET(symbol,psrow,j);
          short  ptyp = OT_GET_short(psrow,j+1);
          if( CFG_KEY(ptyp) )
            parts = xflg
              ? XPT_consT(XPT_newTNode(psym,ptyp,psym),(List(PT_Term))parts)
              : PT_consT(PT_newTNode(psym,ptyp,psym),(PT_Term)parts);
          else
          { PT_Term res, ts, tn;
            if( MAP_defined(visited,psym) ) res = (PT_Term)NULL;
            else
            {
              MAP_dfndom(visited,psym,_);
              res = PTP_nt_transform(tree,psym);
            }
            if( res != (PT_Term)NULL ) 
              parts = xflg ? XPT_consT(res,(List(PT_Term))parts)
                           : PT_consT(res,(PT_Term)parts);
            else
            {
              if( xflg ) freeList((List(PT_Term))parts,PT_delT); 
              else       
              { 
                for( ts = (PT_Term)parts; ts != (PT_Term)NULL; )
                {
                  tn = PT_next(ts); PT_delT(ts); ts = tn;
                }
              }
              j = cnt; break;
            }
          }
        }
        if( j < 1 ) 
        {
          return xflg ? XPT_newNtNode(nt,c_prod,(List(PT_Term))parts)
                      : PT_newNtNode(nt,c_prod,(PT_Term)parts);
        }
      }
    }
  }
  return (PT_Term)NULL;
}

PT_Term PTP_nt_transform_aux(PT_Term tree, symbol nt)
/* performs tree-to-tree transformation on source 'tree';
   uses 'nt' as startsymbol
*/
{ PT_Term res = PTP_nt_transform(tree,nt);
  if (res == (PT_Term)NULL)
    fprintf(stderr,"Transformation failed near (%d,%d) for goal '%s:%s:'\n",
            last_lin,last_col,symbolToString(last_nt),
            symbolToString(last_prod));
  return res;
}

PT_Term PTP_transform(PT_Term tree)
/* performs tree-to-tree transformation on source 'tree' */
{
  return PTP_nt_transform_aux(tree,language);
}
