/* ------------------------------------------------------------------------ */
/*                                                                          */
/* [scn_pre.c]           STYX-konform preprocessing                         */
/*                                                                          */
/* ------------------------------------------------------------------------ */
/* export prefix "SPP_" */

#include "standard.h"
#include "symbols.h"
#include "hmap.h"
#include "scn_base.h"

#include "otab.h"
#include "pathes.h"
#include "sink.h"
#include "gstream.h"

/*!
  <p>The STYX system comes up with a macro preprocessing facility - 
  <br>based on the following macro definition syntax and evaluation process 
  during the lexical analysis:
  <p><table cellspacing=10 noborder>
  <tr>
  <td><b>#include</b> Path</td>
  <td>supports modular grammar design
  <br>The macro will be replaced by the content of source file 'Path' and
      then rescanned.
  </td>
  </tr>
  <tr>
  <td><b>#macro</b> Name ( Parameter , ... ) = Text <b>#end</b></td>
  <td>supports conditional parsing and text replacement
  <br>Parameter and replacement text definition is optional.
  <br>The macro will be collected by the preprocessing module and 
      skipped by the scanner.
  <br>Each macro expression will be replaced according to the previously
      definition and then rescanned.
  </td>
  </tr>
  <tr>
  <td>' Delimiter</td>
  <td>introduces another character as delimiter for the actual macro parameter
  <br>Spaces will be used as default delimiter. 
      To reactive the default behaviour specify ' -.
  <br>The scanner skips these token.
  </td>
  </tr>
  <tr>
  <td><b>#if[n]def</b> Name IfPart <b>#else</b> ElsePart <b>#end</b></td>
  <td>supports conditional parsing
  <br>Dependant on the definition of macro 'Name' the scanner skips the
      'ElsePart' or 'IfPart'. 
  <br>Its possible to introduce environment variables as macros.
  </td>
  </tr>
  </table>
  <br>
  <p>You can use this preprocessing facility if your regular grammar defines
  the above kind of macros in the same way - 
  <br>apart from token names and keywords. 
  ( see below and <a href="styx_cfg.htm"><b>styx</b> reference</a> )
  <br>
  <p>Note: The preprocessing module is not reentrant.<br>
  <br>
*/

/*I-------------------- Macros -------------------------------------------- */

/*!
  <p><b>Identifier for the ( macro ) token names</b>
*/

/*DOC_INTERFACE*/
#define SPP_TOK_IDE         0 // Identifier
#define SPP_TOK_MACDFN      1 // Macro 
#define SPP_TOK_MACSEP      2 // Delimiter
#define SPP_TOK_MACINC      3 // Include
#define SPP_TOK_MACCOND     4 // Condition

/*!
  <p><b>Identifier for macro token keywords</b>
*/

/*DOC_INTERFACE*/
#define SPP_PAT_START       5  // #macro
#define SPP_PAT_END         6  // #end
#define SPP_PAT_IFDEF       7  // #ifdef
#define SPP_PAT_IFNDEF      8  // #ifndef
#define SPP_PAT_ELSE        9  // #else
#define SPP_PAT_INC         10 // #include

/*!
  <p><b>Identifier for include path and character set</b>
*/

/*DOC_INTERFACE*/
#define SPP_INC_PATH        11
#define SPP_INC_CHARSET     12

#define SPP_COND_SCAN       1
#define SPP_COND_SKIP       0
#define SPP_COND_ERROR      -1

#define SPP_ISSPACE(c)      ( (c) == ' '  || (c) == '\t' || \
                              (c) == '\n' || (c) == '\r' )

#define SPP_PREPAR(p)    ( ( pCurPreParMap == (MAP(_,_))NULL ||              \
                         ! MAP_defined(pCurPreParMap,p) ) ? (szPreParIds[p]) \
                         : symbolToString(MAP_apply(symbol,pCurPreParMap,p)) )

// Preprocessing Parameter
static const c_string szPreParIds[] =
{
  "Ide",        // us-ascii or ignore case !!!
  "MacDfn",
  "MacSep",
  "MacInc",
  "MacCond",

  "#macro",
  "#end",
  "#ifdef",
  "#ifndef",
  "#else",
  "#include",

  "PATH",
  ""
};

static MAP(long,symbol) pCurPreParMap = (MAP(long,symbol))NULL;

// type of macro maps
static MAPTY pMapTyp = (MAPTY)NULL;

// macro definition:
// macro name |--> { formal parameter name } macro text 
static MAP(symbol,ROW(symbol)) pMacDfnMap = (MAP(symbol,symbol))NULL;

// evaluate the current token sequence as macro apply:
// ( actual parameter[], macro name )[]
static ROW(_) pCurMacRow = (symbol)NULL;

// macro condition:
// empty row --> scan all
// top = 1   --> scan til #else/#end
// top = 0   --> skip til #else/#end
// top = -1  --> skip all
static ROW(long) pCondRow;

// current macro parameter separator
static char szCurMacSep[3];

// current parameter
static Sink pCurMacPar;


/*I-------------------- Macro Expansion ----------------------------------- */

void SPP_init(MAP(long,symbol) pPreParMap)
/* initializes STYX-konform macro preprocessing 
   'pPreParMap': re-definitions for macro token names,
                 initial macro token keywords, 
                 include path and characterset
*/
{
  pMapTyp       = MAP_newTyp
                  (
                    primCopy,primFree,primEqual,primHash,primCopy,OT_delT
                  );
  pMacDfnMap    = MAP_newMap(pMapTyp);
  pCondRow      = OT_CREATE_ADT(long);
  pCurMacPar    = Sink_open();  
  *szCurMacSep  = '\0';
  pCurPreParMap = pPreParMap;
}

void SPP_addCtxMacro(symbol pMacNam)
/* adds 'pMacNam' as pre-defined macro */
{ ROW(symbol) pMacDfn = OT_CREATE_ADT(symbol);
  MAP_ovrdom(pMacDfnMap,pMacNam,pMacDfn);
  OT_PUSH(stringToSymbol(""),pMacDfn);
}

void SPP_quit(void)
/* terminates STYX-konform macro preprocessing */
{ string szTmp = Sink_close(pCurMacPar);
  pCurMacPar = (Sink)NULL; FreeMem(szTmp);
  if( pCurMacRow != (ROW(symbol))NULL )
  { long lCnt = OT_CNT(pCurMacRow), lI;
    for( lI=0; lI < lCnt; lI += 2 )
      OT_DEL_T(OT_GET(ROW(symbol),pCurMacRow,lI));
    OT_DEL_T(pCurMacRow);  pCurMacRow = (ROW(symbol))NULL;
  }
  MAP_freeMap(pMacDfnMap); pMacDfnMap = (MAP(symbol,symbol))NULL;
  MAP_freeTyp(pMapTyp);    pMapTyp    = (MAPTY)NULL;
  OT_DEL_T(pCondRow);      pCondRow   = (ROW(long))NULL;
}

static symbol SPP_getInclude(string cTokVal)
/* extract include file */
{ string szS = cTokVal + strlen(SPP_PREPAR(SPP_PAT_INC)), szPath;
  symbol pPath;
  while( SPP_ISSPACE(*szS) ) szS++;
  szPath = MkPath(SPP_PREPAR(SPP_INC_PATH),szS,"",'p');
  pPath  = stringToSymbol(szPath);
  FreeMem(szPath);
  return pPath;
}

static void SPP_setCurMacSep(string cTokVal)
{
  if( strlen(cTokVal) == 3 )
  { char cSep = cTokVal[1];
    if( SPP_ISSPACE(cSep) ) *szCurMacSep = '\0';
    else *szCurMacSep = cSep;
  }
  else strcpy(szCurMacSep,cTokVal);
}

static bool SPP_evalCond(string cTokVal)
/* eval macro condition */
{ string szS = cTokVal, szT, szID;
  symbol pID;
  while( !SPP_ISSPACE(*szS) ) szS++;
  while(  SPP_ISSPACE(*szS) ) szS++;
  szID = StrCopy(szS); StrToLower(szID);
  pID  = stringToSymbol(szID);
  FreeMem(szID);
  szT = SPP_PREPAR(SPP_PAT_IFDEF);
  if( !strncmp(cTokVal,szT,strlen(szT)) )
  {
    return MAP_defined(pMacDfnMap,pID);
  }
  else
  {
    return ! MAP_defined(pMacDfnMap,pID);
  }
}

static void SPP_addMacro(string cTokVal)
/* add macro definition */
{ Sink        pSnk    = Sink_open();
  ROW(symbol) pMacDfn = OT_CREATE_ADT(symbol);
  string      szS     = cTokVal + strlen(SPP_PREPAR(SPP_PAT_START)), szT, szVal;
  // extract name
  while( SPP_ISSPACE(*szS) ) szS++;
  while( !SPP_ISSPACE(*szS) && 
         *szS != '(' && *szS != '='  && *szS != '\0' ) 
  {
    Sink_printf(pSnk,"%c",*szS); szS++;
  }
  szVal = Sink_close(pSnk); StrToLower(szVal);
  MAP_ovrdom(pMacDfnMap,stringToSymbol(szVal),pMacDfn);
  FreeMem(szVal); 
  while( SPP_ISSPACE(*szS) ) szS++;
  // extract parameter
  if( *szS == '(' )
  {
    szS++; szT = strstr(szS,")");
    pSnk = Sink_open();
    while( szS != szT )
    {
      if( SPP_ISSPACE(*szS) )
        ;
      else
      if( *szS == ',' )
      {
        szVal = Sink_close(pSnk);
        OT_PUSH(stringToSymbol(szVal),pMacDfn);
        FreeMem(szVal);
        pSnk  = Sink_open(); 
      }
      else Sink_printf(pSnk,"%c",*szS);
      szS++;
    }
    szVal = Sink_close(pSnk);
    OT_PUSH(stringToSymbol(szVal),pMacDfn);
    FreeMem(szVal);
    szS++;
  }
  // extract value
  szT = strstr(szS,"=");
  if( szT != (string)NULL )
  {
    szS   = szT+1;
    szVal = SubStrCopy(szS,strlen(szS)-strlen(SPP_PREPAR(SPP_PAT_END)));
    OT_PUSH(stringToSymbol(szVal),pMacDfn);
    FreeMem(szVal);
  }
  else OT_PUSH(stringToSymbol(""),pMacDfn);
}

static long SPP_nextParameter
            (
              string szMacText, MAP(symbol,long) pParMap
            )
{ MAPIT pItr; symbol pParNam; long lParIdx;
  MAP_forItrAsg(pParNam,lParIdx,pItr,pParMap)
  {
    if( strstr(szMacText,symbolToString(pParNam)) == szMacText )
    {
      MAP_freeItr(pItr); return lParIdx;
    }
  }
  return -1;
}

static symbol SPP_evalMacro(ROW(symbol) pMacDfnRow, ROW(symbol) pMacParRow)
{ MAP(symbol,long) pParMap   = MAP_newPrimMap();
  long             lCnt      = OT_CNT(pMacDfnRow), lIdx, lI;
  symbol           pMacText  = OT_GET(symbol,pMacDfnRow,lCnt-1), pParNam;
  string           szMacText = symbolToString(pMacText), 
                   szS       = szMacText, szT;
  Sink             pSnk      = Sink_open();
  for( lI=0; lI < lCnt-1; ++lI )
  {
    pParNam = OT_GET(symbol,pMacDfnRow,lI);
    if( ! MAP_defined(pParMap,pParNam) ) MAP_dfndom(pParMap,pParNam,lI);
  }
  while( *szS != '\0' )
  {
    lIdx = SPP_nextParameter(szS,pParMap);
    if( lIdx < 0 ) 
    {
      Sink_printf(pSnk,"%c",*szS); ++szS;
    }
    else 
    {
      Sink_printf(pSnk,"%s",symbolToString(OT_GET(symbol,pMacParRow,lIdx))); 
      szS += strlen(symbolToString(OT_GET(symbol,pMacDfnRow,lIdx)));
    }
  }
  szT      = Sink_close(pSnk);
  pMacText = stringToSymbol(szT);
  MAP_freeMap(pParMap); FreeMem(szT);
  return pMacText;
}

static symbol SPP_applyParameter(symbol pTokVal, symbol* pMacNam)
{ symbol      pTopMac    = OT_TOP(symbol,pCurMacRow), pMacVal;
  long        lCnt       = OT_CNT(pCurMacRow);
  ROW(symbol) pMacDfnRow = MAP_apply(ROW(symbol),pMacDfnMap,pTopMac),
              pMacParRow = OT_GET(ROW(symbol),pCurMacRow,lCnt-2);
  OT_PUSH(pTokVal,pMacParRow);
  *pMacNam = pTopMac;
  if( OT_CNT(pMacParRow) == OT_CNT(pMacDfnRow) - 1 )
  {
    pMacVal = SPP_evalMacro(pMacDfnRow, pMacParRow);
    OT_POP(pCurMacRow); OT_POP(pCurMacRow); OT_DEL_T(pMacParRow);
    if( OT_CNT(pCurMacRow) == 0 )
    {
      OT_DEL_T(pCurMacRow); pCurMacRow = (ROW(symbol))NULL;
      return pMacVal;
    }
    else return SPP_applyParameter(pMacVal,pMacNam);
  }
  else return (symbol)NULL;
}

static symbol SPP_applyMacro(symbol pTokVal)
{ ROW(symbol) pMacDfnRow = MAP_apply(ROW(symbol),pMacDfnMap,pTokVal), pTmpRow;
  if( pCurMacRow == (ROW(symbol))NULL )
  {
    if( OT_CNT(pMacDfnRow) == 1 ) return OT_GET(symbol,pMacDfnRow,0);
    else
    {
      pCurMacRow = OT_CREATE_ADT(symbol);
      pTmpRow    = OT_CREATE_ADT(symbol);
      OT_PUSH(pTmpRow,pCurMacRow);
      OT_PUSH(pTokVal,pCurMacRow);
      return (symbol)NULL;
    }
  }
  else
  { symbol pMacNam;
    if( OT_CNT(pMacDfnRow) == 1 ) 
      return SPP_applyParameter(OT_GET(symbol,pMacDfnRow,0),&pMacNam);
    else
    {
      pTmpRow = OT_CREATE_ADT(symbol);
      OT_PUSH(pTmpRow,pCurMacRow);
      OT_PUSH(pTokVal,pCurMacRow);
      return (symbol)NULL;
    }
  }
}

static int SPP_prepare_rescan
           (
             Scn_Stream pStream, int nScnAction, symbol pMacNam, symbol* cMacVal
           )
/* prepare rescan action */
{ string szCharSet = (string)NULL;
  if( ( nScnAction == SCN_FLG_RescanString  && 
        ! Stream_unicode(pStream) )         ||
      ( nScnAction == SCN_FLG_RescanExtFile &&   
        strlen(szCharSet = SPP_PREPAR(SPP_INC_CHARSET)) == 0 ) )
  { 
    return nScnAction;
  }
  else
  { Scn_Stream_Itr pItr     = (Scn_Stream_Itr)NULL;
    GStream_T      pGS      = (GStream_T)NULL;
    c_string       szMacVal = symbolToString(*cMacVal),
                   szGSName = szMacVal;
    if( nScnAction == SCN_FLG_RescanExtFile )
    { FILE* pFP = OpnFile(szMacVal,"rb");
      pGS = GS_stream_file_new(pFP, szCharSet, C_True);
    } 
    else
    if( nScnAction == SCN_FLG_RescanString )
    {
      pGS = GS_stream_string_new(szMacVal,strlen(szMacVal)+1,CS_ID_MBYTE);
      if( pMacNam != (symbol)NULL )
        szGSName = symbolToString(pMacNam);
    } 
    assert1(pGS != (GStream_T)NULL, "creation of stream '%s' failed",szMacVal);
    pItr = Stream_Itr_new
           (
             GS_stream_get_wcrc, GS_fun_destruct(pGS), GS_ucs4_to_utf8, 
             GS_fun_seek(pGS), pGS, szGSName
           );
    *cMacVal = (symbol)pItr;
    return SCN_FLG_RescanChrItr;
  }
}

int SPP_premac
    (
      Scn_Stream pStream, string cTokNam, string cTokVal, symbol* cMacVal
    )
/* preprocesses current token of scan stream 'pStream'
   'cTokNam': name of separated token
   'cTokVal': value of separated token
   'cMacVal': preprocessing result ( see [scn_base] )
   ( cTokNam, cTokVal: single byte or utf-8 characterset ) 
*/
{ long lCnt = OT_CNT(pCondRow);
  if( !strcmp(cTokNam,SPP_PREPAR(SPP_TOK_MACCOND)) )
  { 
    if( !strcmp(cTokVal,SPP_PREPAR(SPP_PAT_END)) )
    {
      if( lCnt == 0 ) OT_PUSH(SPP_COND_ERROR,pCondRow);
      else            OT_POP(pCondRow);
    }
    else
    if( !strcmp(cTokVal,SPP_PREPAR(SPP_PAT_ELSE)) )
    {
      if( lCnt == 0 ) OT_PUSH(SPP_COND_ERROR,pCondRow);
      else
      {
        if( OT_TOP(long,pCondRow) == SPP_COND_SKIP )
          OT_UPD(long,pCondRow,lCnt-1,SPP_COND_SCAN);
        else 
        if( OT_TOP(long,pCondRow) == SPP_COND_SCAN )
          OT_UPD(long,pCondRow,lCnt-1,SPP_COND_SKIP);
      }
    }
    else
    if( lCnt == 0 || OT_TOP(long,pCondRow) == SPP_COND_SCAN )
    {
      if( SPP_evalCond(cTokVal) ) 
        OT_PUSH(SPP_COND_SCAN,pCondRow);
      else OT_PUSH(SPP_COND_SKIP,pCondRow);
    }
    else OT_PUSH(SPP_COND_ERROR,pCondRow);
    *cMacVal = (symbol)NULL;
    return SCN_FLG_RescanNone;
  }
  else
  {
    if( lCnt > 0 &&
        ( OT_TOP(long,pCondRow) == SPP_COND_ERROR ||
          OT_TOP(long,pCondRow) == SPP_COND_SKIP   ) )
    {
      *cMacVal = (symbol)NULL;
      return SCN_FLG_RescanNone;
    }
    else
    if( !strcmp(cTokNam,SPP_PREPAR(SPP_TOK_MACINC)) )
    { 
      *cMacVal = SPP_getInclude(cTokVal);
      return SPP_prepare_rescan(pStream,SCN_FLG_RescanExtFile,(symbol)NULL,cMacVal);
    }
    else
    if( !strcmp(cTokNam,SPP_PREPAR(SPP_TOK_MACDFN)) )
    { 
      SPP_addMacro(cTokVal);
      *cMacVal = (symbol)NULL;
      return SCN_FLG_RescanNone;
    }
    else
    { symbol pValSym    = stringToSymbol(cTokVal), 
             pLowValSym = pValSym, 
             pMacNam    = (symbol)NULL;
      if( !strcmp(cTokNam,SPP_PREPAR(SPP_TOK_IDE)) && !Stream_cica(pStream) )
      { c_string szLowTokVal = StrCopy(cTokVal);
        StrToLower(szLowTokVal);
        pLowValSym = stringToSymbol(szLowTokVal);
        FreeMem(szLowTokVal);
      } 
      if( !strcmp(cTokNam,SPP_PREPAR(SPP_TOK_IDE)) && 
          MAP_defined(pMacDfnMap,pLowValSym)        )
      { 
        *cMacVal = SPP_applyMacro(pLowValSym);
        return ( *cMacVal != (symbol)NULL )
               ? SPP_prepare_rescan(pStream,SCN_FLG_RescanString,pLowValSym,cMacVal) 
               : SCN_FLG_RescanNone;
      }
      else
      if( !strcmp(cTokNam,SPP_PREPAR(SPP_TOK_MACSEP)) && 
          ( pCurMacRow == (ROW(symbol))NULL || strcmp(cTokVal,szCurMacSep) ) )
      {
        SPP_setCurMacSep(cTokVal);
        *cMacVal = (symbol)NULL;
        return SCN_FLG_RescanNone;
      }
      else
      if( pCurMacRow != (ROW(symbol))NULL )
      {
        if( strlen(szCurMacSep) == 0 && Sink_empty(pCurMacPar) ) 
        {
          *cMacVal = SPP_applyParameter(pValSym,&pMacNam);
          return ( *cMacVal != (symbol)NULL )
                 ? SPP_prepare_rescan(pStream,SCN_FLG_RescanString,pMacNam,cMacVal) 
                 : SCN_FLG_RescanNone;
        }
        else
        {
          if( strcmp(cTokNam,SPP_PREPAR(SPP_TOK_MACSEP)) && 
              strcmp(cTokVal,szCurMacSep)                 )
            Sink_printf(pCurMacPar,"%s",cTokVal);
          if( strlen(szCurMacSep) == 0                    ||
              !strcmp(cTokNam,SPP_PREPAR(SPP_TOK_MACSEP)) || 
              !strcmp(cTokVal,szCurMacSep)                 )
          { string szPar = Sink_close(pCurMacPar);
            *cMacVal = SPP_applyParameter(stringToSymbol(szPar),&pMacNam);
            FreeMem(szPar); pCurMacPar = Sink_open();
            return ( *cMacVal != (symbol)NULL )
                   ? SPP_prepare_rescan(pStream,SCN_FLG_RescanString,pMacNam,cMacVal) 
                   : SCN_FLG_RescanNone;
          } 
          else
          {
            *cMacVal = (symbol)NULL;
            return SCN_FLG_RescanNone;
          }
        }
      }
      *cMacVal = pValSym;
      return SCN_FLG_RescanNone;
    }
  }
}

