/* ------------------------------------------------------------------------ */
/*                                                                          */
/* [symbols.c]                    Type: Symbol                              */
/*                                                                          */
/* Copyright (c) 1993 by D\olle, Manns                                      */
/* ------------------------------------------------------------------------ */

#include "standard.h"

#include "literal.h"
#include "hmap.h"


/*!! 
    This module implements the symbol type for a unique representation of strings 
    and the symbol table in which they are stored. One intention is the reduction 
    of place- and time-complexity when dealing with many identical strings.
    
    Spezification
    1) forall String s : strcmp(s,symbolToString(stringToSymbol(s))) == 0.
    2) forall Symbol s : equalSymbols(s,stringToSymbol(symbolToString(s)).
    3) equalSymbols is an equality.
*/

/*I------------------------------ The Type -------------------------------- */

ExternalType( symbol ) /* Abstract symbol type */
{
  long key;
  char str[1];
};

static MAP(string,symbol) symtab = (MAP(string,symbol)) NULL;

/*I----------------------------- Init & Quit ------------------------------ */

void initSymbols(void)
/* initializes the symbol table */
{ MAPTY symtyp;
  assert0(symtab == (MAP(string,symbol)) NULL, "symbol table allready initalized");
  symtyp = MAP_newTyp(primCopy, FreeMem, strEqual, strHash, primCopy, FreeMem);
  symtab = MAP_newMap(symtyp);
}

void freeSymbols(void)
/* releases the symbol table */
{ MAPTY symtyp;
  assert0(symtab != (MAP(string,symbol)) NULL, "symbol table not initalized");
  symtyp = MAP_MapTyp(symtab);
  MAP_freeMap(symtab); MAP_freeTyp(symtyp);
  symtab = (MAP(string,symbol)) NULL;
}

/*I----------------------------- Conversion ------------------------------- */

#define MAGIC 0x7654321L

symbol stringToSymbol(string str)
/* converts string 'str' into a symbol */
{ StdCPtr sem = sysbase_sem();
  symbol  sym = (symbol)NULL;
  assert0(symtab != (MAP(string,symbol)) NULL, "symbol table not initalized");
  if( sem == (StdCPtr)NULL || waitSem(sem) == 1 )
  {
    if (!MAP_defined(symtab, str))
    { symbol s = (symbol)NewMem(SizeOf(symbol)+strlen(str));
      s->key = /*strHash(str)*/ MAGIC;
      strcpy(s->str,str);
      MAP_dfndom(symtab, StrCopy(str), s);
      sym = s;
    }
    else sym = MAP_apply(symbol,symtab,str);
    if( sem != (StdCPtr)NULL ) postSem(sem,1);
  }
  return sym;
}

symbol lstringToSymbol(string str)
/* converts ( lower case ) string 'str' into a symbol */
{ string lstr;
  symbol lsym;
  assert0(symtab != (MAP(string,symbol)) NULL, "symbol table not initalized");
  lstr = StrCopy(str); StrToLower(lstr);
  lsym = stringToSymbol(lstr);
  FreeMem(lstr);
  return lsym;
}

symbol ustringToSymbol(string str)
/* converts ( upper case ) string 'str' into a symbol */
{ string ustr;
  symbol usym;
  assert0(symtab != (MAP(string,symbol)) NULL, "symbol table not initalized");
  ustr = StrCopy(str); StrToLower(ustr);
  usym = stringToSymbol(ustr);
  FreeMem(ustr);
  return usym;
}

string symbolToString(symbol sym)
/* string corresponding to symbol 'sym' */
{
  assert(sym->key==MAGIC,"symbolToString : non-symbol");
  return sym->str;
}

/*I------------------------------ Predicates ------------------------------ */

bool equalSymbols(symbol a, symbol b)
/* whether two symbols 'a' and 'b' are equal */
{
  return a == b;
}

/*I----------------------------- Debugging -------------------------------- */

void printSymbols(int indent)
/* prints the symbol table to stdout */
{ MAPIT itr; symbol e;
  assert0(symtab != (MAP(string,symbol)) NULL, "symbol table not initalized");
  fprintf(STDOUT,"Symbols /* count = %ld */",MAP_count(symtab));
  NL; fprintf(STDOUT,"{"); NL;
  MAP_forItr(e, itr, symtab)
  { string s = LIT_c_str_lit(e->str);
    fprintf(STDOUT,"  Symbol (%08lx) ",e->key);
    //fprintf(STDOUT,"%s",e->str);
    fprintf(STDOUT,"%s",s);
    if (!MAP_emptyItr(itr))
      fprintf(STDOUT,",");
    NL;
    FreeMem(s);
  }
  fprintf(STDOUT,"}");
  NL; NL;
}

int countSymbols(void)
/* number of symbols in the symbol table */
{
  return MAP_count(symtab);
}
