/* ------------------------------------------------------------------------ */
/*                                                                          */
/* [prs_gen.c]              LALR(1) Parser Generator                        */
/*                                                                          */
/* Copyright (c) 1993 by D\olle, Manns                                      */
/* ------------------------------------------------------------------------ */

#include "prs_dfn.h"

/*!
  <p>The module [prs_gen] performs two main tasks.

  <p>1) A set of functions will be used to dynamically create a BNF-like context
        free grammar definition. ( reentrant )
        <br>It is possible to define multiple startsymbols and comment tokens.

  <p>2) The latter function create the corresponding parse table and nonterminal 
        classes. 
        The parse table is the input for the <a href="prs.htm">parse function</a>.
        <br>This creation process is not reentrant. You must synchronize the access
        to the function within threads.     
*/

/*!!
  The following rules define the abstract syntax. On this base the nonterminal
  classes are created.

  A) well-formed productions:
  1. let X :nil  :            <0 members>
  2. let X :cons : Y Z        <2 members: Y = member && Z = nonterminal>
  3. let X :ign#*: Y          <1 nonterminal>
  4. let X :name : X1 .. Xn   <n >= 0 && name =/= { ign1, nil, cons }>

  B) construction of the token/nonterminal classes:
  1. X <=> X                          reflexiv 
  2. X <=> Y --> Y <=> X              symmetric
  3. X <=> Y && Y <=> Z --> X <=> Z   transitiv
  4. let X :ign1: Y   --> X <=> Y
  5. let X :cons: Y Z --> X <=> Z  
  6. X <=> Y && let X :idx: X1 .. Xn && let Y :idy: Y1 .. Ym && idx = idy
     --> n = m && forall i, 1 <= i <= n: Type(Xi) = Type(Yi) && Xi <=> Yi,
                                         where Type(Z) = { token, nonterminal }
  7. all tokens are equivalent.

  C) token/nonterminal classes:
     [X] = { Y | Y <=> X }
     class representants:
     - tokens:       "Tok"
     - startsymbols: language name
     - nonterminals: less nonterminal name according the lexical order

  D) correctness:
  1. X <=> Y --> Type(X) = Type(Y), where Type(Z) = { token, nonterminal }
  2. let X^ :id: a && let X^ :id: b --> a <=> b
  3. let X^ :nil: a || let X^ :cons: b
     --> not exists P: P = let X^ :id: c && id =/= { ign1, nil, cons }
  [ 1,2: checked during construction ]

  E) abstract context free grammar:
     NT |--> NT^
     T  |--> T^ ( NT^ T^ are the class representants )

     for all "normal" productions there will be one interface function 
     of type 'bool' which returns whether the argument term represents 
     a production of this kind and in the positive case all required members.
*/

/*D
  Aenderungen fuer einen reflexiven Parser

   Nonterminals waeren zugleich Terminals.
   1) Die First-Sets der Nonterminals muessten um die jeweilige
      Nonterminal-Huelle erweitert werden.
   2) Nonterminals werden moegliche Lookaheads.
      ( Set[KFGSYMCNT+STARTCNT] )
   3) Die GoTo-Tabellen entfallen, und die Shift/Reduktions-Tabellen
      beziehen sich auch auf Nonterminals.
*/

#include "olist.h"
#include "otab.h"
#include "binset.h"
#include "prs_imp.h"

/*==================== Types / Constants / Globals ===========================*/

/*I------------------------------ Types ------------------------------------- */

AbstractType(PLR_Cfg); /* Abstract context free grammar type */

typedef struct _PSDFN {
/* production symbol ( definition interface ) */
  bool istoken; /* symbol is a token */
  int  symbol;  /* symbol            */
  int  sRow;    /* line position     */
  int  sCol;    /* col position      */
} PSDFN;

typedef struct _PRDFN {
/* production ( definition interface ) */
  string pname;   /* production       */
  int    method;  /* reduction method */
  int    pid;     /* nonterminal      */
  OT_Tab psymbol; /* symbols          */
} PRDFN;

typedef struct _KFGDFN {
/* context free grammar ( definition interface ) */
  string language; /* language            */
  OT_Tab Token;    /* token               */
  OT_Tab TkIdx;    /* token indexes       */
  OT_Tab NtIdx;    /* nonterminal indexes */
  OT_Tab NonTerm;  /* nonterminals        */
  OT_Tab StartNt;  /* startsymbols        */
  OT_Tab TokKind;  /* token kind          */
  OT_Tab ErrorNt;  /* error recovery nt's */
  OT_Tab Prod;     /* productions         */
} KFGDFN;

typedef struct _KFGNT {
/* nonterminal */
  int    altcnt; /* productions                      */
  int    fstalt; /* first produktion                 */
  OT_Tab states; /* occurrence in StateTab as symbol */
  BS_Set first;  /* FIRST set                        */
} KFGNT;

typedef struct _PSTATE {
/* parser state */
  int    symbol; /* with symbol marked vertice or -1 ( start state ) */
  OL_Lst states; /* follow states                                    */
  OL_Lst kern;   /* kernel elements                                  */
  OT_Tab looks;  /* lookaheads                                       */
} PSTATE;

/*D
   Darstellung der Kernelemente

   Ein (Kern-)Element wird durch einen Index in eine
   vektorisierte Matrix [ProdCnt, MaxSymCnt+1] dargestellt,
   wobei die 1. Dimension die Nr. der Produktion und
   die 2. Dimension die Punkt-Position angibt.
   ( Position vor dem durch die 2. Dimension numerierten
     Symbol oder MaxSymCnt+1 zur Kennzeichnung einer
     Reduktionsproduktion )
*/

/* ------------------------ Constants and Globals --------------------------- */

static bool profile;       /* Profile Printing                    */
static int ProdCnt;        /* Anzahl der Grammatik-Produktionen   */
static int MaxSymCnt;      /* max. Anzahl Produktionssymbole      */
static int MaxSymNamLen;   /* max. Laenge der Symbol Namen        */
static int conflictCnt;    /* Anzahl Konflikte                    */

static OT_Tab* NtEquiv;    /* Equivalence Tables of Nonterminals  */

static KFGHEAD* Kfg;       /* Grammatik-Kopfdaten                 */
static KFGNT* NonTermTab;  /* Nonterminals der Grammatik          */
static KFGPROD* ProdTab;   /* Produktionen der Grammatik          */
static OT_Tab GoToKern;    /* Kerne der Folgezustaende            */
                           /* ( Tabelle von Kernelement-Listen )  */
static OT_Tab KernTab;     /* Indizes Kernelemente                */
static OT_Tab StateTab;    /* Parser-Zustaende                    */
static BS_Set LambdaSet;   /* LAMBDA-Set { Nt *=> lambda }        */
static BS_Set TokenSet;    /* FIRST-Set of current Symbol-List    */
static BS_Set NonTermSet;  /* Set for marking Nonterminals        */
static BS_Set NtRelSet;    /* Nt-LALR(0)-Huelle                   */
static BS_Set StateSet;    /* Set for Propagation of Lookaheads   */
static BS_Set* PLaheads;   /* Lookahead-Sets fuer die aktuellen   */
                           /* Huellen-Elemente                    */

#define MAX_OBJECTS       10000

#define MAX_PR_ICOLS      10
#define MAX_PR_SCOLS      5

#define PLR_C_IMP(cfg)    ( (KFGDFN*)(cfg) )

#define NTIDX(nt)         ( ( nt ) - Kfg->TkCnt )
#define NTSYMBOL(idx)     ( ( idx ) + Kfg->TkCnt )
#define NTCLASS(idx)      ( Kfg->NtClass[idx] )
#define NTEQUIV(idx)      ( NtEquiv[idx] )
#define FIRSTNTPROD(nt)   ( NonTermTab[NTIDX(nt)].fstalt )
#define LASTNTPROD(nt)    ( FIRSTNTPROD(nt) + NonTermTab[NTIDX(nt)].altcnt - 1 )
#define FIRSTSET(nt)      ( NonTermTab[NTIDX(nt)].first )
#define NONTERMCNT        ( Kfg->NtCnt + Kfg->StartCnt )
#define TERMCNT           ( Kfg->TkCnt + Kfg->StartCnt )
#define KFGSYMCNT         ( Kfg->TkCnt + Kfg->NtCnt )
#define ACCEPTTK(i)       ( ( i ) + Kfg->TkCnt )
#define TKACCEPT(tk)      ( ( tk ) - Kfg->TkCnt )
#define NTACCEPT(nt)      ( ( nt ) - Kfg->NtCnt )

#define KERNMATRIX        ( ProdCnt * ( MaxSymCnt + 1 ) )
#define KERNELEMENT(li,r) ( ( ( li ) * ( MaxSymCnt + 1 ) ) + ( r ) )
                          /* li = Prod.index   0 .. ProdCnt-1   */
                          /* r  = Pkt.position 1 .. MaxSymCnt+1 */
#define PRODUCTION(e)     ( (OT_GET(unsigned long,KernTab,(e)-1)) >> 16 )
#define POSITION(e)       ( (OT_GET(unsigned long,KernTab,(e)-1)) & 0xFFFF )

#define NEW_PSYMTAB       OT_CREATE(PSDFN,cpyPsym,delPsym,primEqual)
#define INS_PSYM(p,t)     OT_T_INS(t,p)
#define GET_PSYM(t,i)     OT_GET(PSDFN*,t,i)

#define NEW_PRODTAB       OT_CREATE(PSDFN,cpyProd,delProd,primEqual)
#define INS_PROD(p,t)     OT_T_INS(t,p)
#define GET_PROD(t,i)     OT_GET(PRDFN*,t,i)
#define SRT_PRODS(t)      OT_SORT(t,cmpProd)
#define FND_PROD(t,ti,k)  OT_B_FIND(t,ti,(k),cmpProd)

#define NEW_STATETAB      OT_CREATE(PSTATE,cpyState,delState,primEqual)
#define INS_STATE(s,t)    OT_T_INS(t,s)
#define GET_STATE(t,i)    OT_GET(PSTATE*,t,i)

#define NEW_KLISTTAB      OT_CREATE(OL_Lst,primCopy,OL_delL,OL_equal)
#define INS_KLIST(l,t)    OT_T_INS(t,l)
#define GET_KLIST(t,i)    OT_GET(OL_Lst,t,i)

#define NEW_LOOKTAB       OT_CREATE(BS_Set,primCopy,delLook,BS_equal)
#define INS_LOOK(l,t)     OT_T_INS(t,l)
#define INSN_LOOK(l,t,n)  OT_P_INS(t,l,n)
#define GET_LOOK(t,i)     OT_GET(BS_Set,t,i)

#define NEW_INTTAB        OT_CREATE_ADT(int)
#define INS_INTTE(e,t)    OT_T_INS(t,e)
#define UPD_INTTE(e,t,i)  OT_UPD(int,t,i,e)
#define GET_INTTE(t,i)    OT_GET(int,t,i)
#define INTTE_ARRAY(t)    OT_COPY_V(int*,t,sizeof(int),cpyInt)

#define NEW_STRTAB        OT_CREATE(string,StrCopy,FreeMem,strEqual)
#define INS_STRTE(e,t)    OT_T_INS(t,e)
#define GET_STRTE(t,i)    OT_GET(string,t,i)
#define SRT_STRTES(t)     OT_SORT(t,cmpStr)
#define FND_STRTE(t,ti,k) OT_B_FIND(t,ti,(k),strcmp)

#define NEW_INTLIST       OL_CREATE_ADT(INT)
#define INSC_INTLE(e,l)   OL_C_INS(INT,l,e)
#define INST_INTLE(e,l)   OL_T_INS(INT,l,e)
#define INSS_INTLE(e,l)   OL_S_INS(INT,l,e,primLessEqual)
#define FND_INTLE(e,l)    OL_FIND(l,e,primGreatEqual)
#define FST_INTLE(l)      OL_FIRST(INT,l)
#define NXT_INTLE(l)      OL_NEXT(INT,l)
#define CURR_INTLE(l)     OL_CURR(INT,l)
#define PRT_INTLES(l,n)   OL_PRINT(l,printINTObj,MAX_PR_ICOLS,(n))

#define ACT_ERROR         ( ProdCnt + 1 )
#define ACT_ACCEPT        ( ProdCnt + 2 )
#define PROD_OFFSET       ( ProdCnt + TERMCNT )

#define PROD_NORMAL       -3
#define PROD_FALSE        -2
#define PROD_NIL          -1

/*============================= Auxiliary ====================================*/

/* Parameter Check: CALL Function */

static int cmpStr(OT_Obj left, OT_Obj right)
/* left key - right key */
{
  return( strcmp((string)left,(string)right) );
}

static OT_Obj cpyTObj(OT_Obj obj, int size)
/* copy obj */
{
  return( (OT_Obj)BytCopy(obj,size) );
}

static OT_Obj cpyPsym(OT_Obj psym)
/* copy production symbol */
{
  return( cpyTObj(psym,sizeof(PSDFN)) );
}

static OT_Obj cpyProd(OT_Obj prod)
/* copy production */
{
  return( cpyTObj(prod,sizeof(PRDFN)) );
}

static OT_Obj cpyState(OT_Obj state)
/* copy state */
{
  return( cpyTObj(state,sizeof(PSTATE)) );
}

static int cmpProd(OT_Obj left, OT_Obj right)
/* left key - right key */
{
  return( ((PRDFN*)left)->pid - ((PRDFN*)right)->pid );
}

static int* cpyInt(OT_Obj obj)
/* copy an int object */
{ int* ptr = (int*)NewMem(sizeof(int));

  *ptr = (int)obj; return( ptr );
}

static bool addNewMembers(BS_Set dst, BS_Set left, BS_Set right)
/* ! ( right <= left ) ==> dst = left + right */
{
  if( ! BS_subset(right,left) )
  {
    BS_union(dst,left,right); return( True );
  }
  return( False );
}

static INT nextProdElement(INT element)
/* naechstes Kernelement ( innerhalb einer Produktion ) */
{ int idx = PRODUCTION(element) - 1, pos = POSITION(element);
  return
  ( 
    ( pos < ProdTab[idx].symcnt ) 
    ? KERNELEMENT(idx,pos+1)
    : KERNELEMENT(idx,MaxSymCnt+1) 
  );
}

static INT prevProdElement(INT element)
/* vorhergehendes Kernelement ( innerhalb einer Produktion ) */
{ int idx = PRODUCTION(element) - 1, pos = POSITION(element);
  return
  (
    ( pos == MaxSymCnt+1 )
    ? KERNELEMENT(idx,ProdTab[idx].symcnt)
    : KERNELEMENT(idx,pos-1) 
  );
}

static int symType(int SymId)
/* Type(SymId) */
{
  return
  ( 
    ( SymId >= Kfg->TkCnt ) ? PLR_TYP_NTM : Kfg->TokKind[SymId] 
  );
}

static bool isToken(int SymId)
/* SymId ist ein "User"-Token */
{
  return( ( SymId >= 0 && SymId < Kfg->TkCnt ) ? True : False );
}

static bool isNonTerm(int SymId)
/* SymId ist ein "User"-Nonterminal */
{
  return( ( SymId >= Kfg->TkCnt && SymId < KFGSYMCNT ) ? True : False );
}

static int cntProdMembers(KFGPROD* prod)
/* number of tokens and nonterminals of 'prod' */
{ int cnt, typ, i;
  for( cnt=0, i=0; i < prod->symcnt; ++i )
  {
    typ  = symType(prod->symbols[i]);
    cnt += CFG_TOK_NTM(typ);
  }
  return( cnt );
}

static int nextProdMember(KFGPROD* prod, int startidx)
/* next token/nonterminal of 'prod' */
{ int typ, i;
  for( i=startidx; i < prod->symcnt; ++i )
  {
    typ = symType(prod->symbols[i]);
    if( CFG_TOK_NTM(typ) ) return( i );
  }
  return( prod->symcnt );
}

static void forallNtClasses(void (*act)(int ntclass))
/* performs 'act' on all nonterminal classes */
{ int i;
  for( i=0; i < Kfg->NtCnt; ++i ) if( i == NTCLASS(i) ) (*act)(i);
}

static void forallEquivNts
            (
              int ntclass, void (*act)(int ntidx, int data), int data
            )
/* performs 'act' on all equivalent nonterminals of 'ntclass' */
{ int cnt = OT_CNT(NTEQUIV(ntclass)), i;
  for( i=0; i < cnt; ++i ) (*act)(OT_GET(int,NTEQUIV(ntclass),i),data);
}

/*====================== Creation of the Parse Tables ========================*/

/* Parameter Check: CALL Function */

/* ------------------------ Nonterminal Classes ----------------------------- */

static int wellProd(KFGPROD* prod)
/* whether 'prod' is well-formed */
{ int res = PROD_FALSE, typ, cnt, i;
  if( nilProd(prod->pname) )
  {
    if( cntProdMembers(prod) ) return( PROD_FALSE );
    else                       return( PROD_NIL );
  }
  else
  {
    if( consProd(prod->pname) )
    {
      for( cnt=0, i=0; i < prod->symcnt; ++i )
      {
        typ  = symType(prod->symbols[i]);
        cnt += CFG_TOK_NTM(typ);
        if( cnt == 2 && CFG_NTM(typ) ) res = prod->symbols[i];
      }
      if( cnt != 2 ) return( PROD_FALSE );
      else           return( res );
    }
    else
    {
      if( ignProd(prod->pname) )
      {
        for( cnt=0, i=0; i < prod->symcnt; ++i )
        {
          typ  = symType(prod->symbols[i]);
          cnt += CFG_TOK_NTM(typ);
          if( cnt == 1 && CFG_NTM(typ) ) res = prod->symbols[i];
        }
        if( cnt != 1 ) return( PROD_FALSE );
        else           return( res );
      }
      else return( PROD_NORMAL );
    }
  }
}

static void checkListProds(int ntclass)
/* checks condition 3 for 'ntclass' */
{ bool isList = False; bool isOther = False;
  int cnt = OT_CNT(NTEQUIV(ntclass)), i; 
  for( i=0; i < cnt; ++i )
  { int nti = OT_GET(int,NTEQUIV(ntclass),i), j;
    for( j =  FIRSTNTPROD(NTSYMBOL(nti)); j <= LASTNTPROD(NTSYMBOL(nti)); ++j )
    { KFGPROD* prod = &ProdTab[j];
      if( nilProd(prod->pname) || consProd(prod->pname) ) isList = True;
      else
        if( ! ignProd(prod->pname) ) isOther = True;
      bug1
      ( 
        ! ( isList && isOther ), 
        "[%s] consists of list/non list productions",
        Kfg->SNames[prod->id]
      );
    }
  }
}

static void changeNtClass(int ntidx, int newClass)
/* Class(ntidx) = newClass */
{
  NTCLASS(ntidx) = newClass;
}

static bool setEquivNts(int ntidx1, int ntidx2)
/* sets nonterminals 'ntidx1' and 'ntidx2' equivalent */
{ int ntc1 = NTCLASS(ntidx1), ntc2 = NTCLASS(ntidx2);
  if( ntc1 == ntc2 ) return( False );
  else
  { 
    if( strcmp(Kfg->SNames[NTSYMBOL(ntc2)],Kfg->SNames[NTSYMBOL(ntc1)]) < 0 )
    {
      forallEquivNts(ntc1,changeNtClass,ntc2);
      OT_APPEND(NTEQUIV(ntc2),NTEQUIV(ntc1));
      OT_DEL_T(NTEQUIV(ntc1));
    }
    else
    {
      forallEquivNts(ntc2,changeNtClass,ntc1);
      OT_APPEND(NTEQUIV(ntc1),NTEQUIV(ntc2));
      OT_DEL_T(NTEQUIV(ntc2));
    }
    return( True );
  }
}

static bool setEquivProdNts(int ntidx1, int ntidx2)
/* determines equivalent production nonterminals of 'ntidx1' and 'ntidx2' */
{ bool change = False;
  if( NTCLASS(ntidx1) == NTCLASS(ntidx2) )
  { int i, j;
    for( i =  FIRSTNTPROD(NTSYMBOL(ntidx1)); 
         i <= LASTNTPROD(NTSYMBOL(ntidx1)); ++i )
    {
        for( j =  FIRSTNTPROD(NTSYMBOL(ntidx2)); 
             j <= LASTNTPROD(NTSYMBOL(ntidx2)); ++j )
        { KFGPROD pi = ProdTab[i], pj = ProdTab[j];
          if( ! strcmp(pi.pname,pj.pname) ) 
          { int k = nextProdMember(&pi,0), l = nextProdMember(&pj,0);
            bug
            ( 
              cntProdMembers(&pi) == cntProdMembers(&pj),
              "symbols of productions '%s::%s', '%s::%s' not equivalent", 
              Kfg->SNames[pi.id],pi.pname,Kfg->SNames[pj.id],pj.pname
            );
            while( k < pi.symcnt && l < pj.symcnt )
            { int ksym = pi.symbols[k], lsym = pj.symbols[l], typ; 
              bug
              ( 
                ( typ = symType(ksym) ) == symType(lsym),
                "token/nonterminal conflict in productions '%s::%s', '%s::%s'", 
                Kfg->SNames[pi.id],pi.pname,Kfg->SNames[pj.id],pj.pname
              );
              if( CFG_NTM(typ) && setEquivNts(NTIDX(ksym),NTIDX(lsym)) ) 
                change = True;
              k = nextProdMember(&pi,k+1); l = nextProdMember(&pj,l+1);
            }
          }
        }
    }
  }
  return( change );
}

static bool pass1NtPartition(KFGPROD* prod)
/* changes the nonterminal partition according 'prod' */
{ int res;
  switch( res = wellProd(prod) )
  {
    case PROD_NORMAL:
    case PROD_NIL:
      return( False );
    case PROD_FALSE:
      bug
      (
        False,"production '%s::%s' not well-formed",
        Kfg->SNames[prod->id],prod->pname
      ); 
      return( False );
    default:
      return( setEquivNts(NTIDX(prod->id),NTIDX(res)) );
  }
}

static void pass2NtPartition(void)
/* completes & checks the definition of the nonterminal partition */
{ bool change = True;
  while( change )
  { int i;
    change = False;
    for( i=0; i < Kfg->NtCnt; ++i ) 
    { OT_Tab cls; int cnt, j, k;
      if( i == NTCLASS(i) && ( cnt = OT_CNT(cls = NTEQUIV(i)) ) > 1 )
        for( j=0; j < cnt; ++j ) for( k=j+1; k < cnt; ++k )
          if( setEquivProdNts( OT_GET(int,cls,j),OT_GET(int,cls,k)) )
            change = True;
    }
  }
  forallNtClasses(checkListProds);
}

/* ------------------------ Release Globals --------------------------------- */

static void delEquivNts(int ntclass)
/* deletes the equivalence table of 'ntclass' */
{
  OT_DEL_T(NTEQUIV(ntclass));
}

static void delLook(StdCPtr SetEl)
/* Loeschen eines Lookahead-Sets */
{
  BS_delS((BS_Set)SetEl);
}

static void delPsym(StdCPtr PsymEl)
/* Loeschen eines Produktionssymbol */
{
  FreeMem((PSDFN*)PsymEl);
}

static void delProd(StdCPtr ProdEl)
/* Loeschen einer Produktion */
{
  FreeMem(((PRDFN*)ProdEl)->pname);
  OT_DEL_T(((PRDFN*)ProdEl)->psymbol);
  FreeMem((PRDFN*)ProdEl);
}

static void delState(StdCPtr StateEl)
/* Loeschen eines Parserzustands */
{
  OL_DEL_L(((PSTATE*)StateEl)->states);
  OL_DEL_L(((PSTATE*)StateEl)->kern);
  OT_DEL_T(((PSTATE*)StateEl)->looks);
  FreeMem((PSTATE*)StateEl);
}

static void delCreateGlobals(void)
/* Loeschen nicht mehr benoetigter Grammatik-Objekte */
{ int i;

  forallNtClasses(delEquivNts);
  FreeMem(NtEquiv);
  BS_delS(LambdaSet);
  BS_delS(TokenSet);
  BS_delS(NonTermSet);
  BS_delS(NtRelSet);
  BS_delS(StateSet);
  for( i=0; i < NONTERMCNT; ++i ) BS_delS(NonTermTab[i].first);
  for( i=0; i < NONTERMCNT; ++i ) OT_DEL_T(NonTermTab[i].states);
  FreeMem(NonTermTab);
  OT_DEL_T(KernTab);
  OT_DEL_T(GoToKern);
  OT_DEL_T(StateTab);
}

/* ----------------------- Grammar Description ------------------------------ */

static void setKfgGlobals(PLR_Cfg Cfg)
/* Initialisierung der globalen Variablen/Grammatik-Kopfdaten */
/* ( inhaltliche Pruefung bei der Grammatik-Definition )      */
{ int i;

  BUG_RNG0(OT_CNT(PLR_C_IMP(Cfg)->Token),MAX_OBJECTS);
  BUG_RNG1(OT_CNT(PLR_C_IMP(Cfg)->NonTerm),MAX_OBJECTS);
  BUG_RNG1(OT_CNT(PLR_C_IMP(Cfg)->StartNt),MAX_OBJECTS);
  MaxSymNamLen  = 0;
  Kfg           = (KFGHEAD*)NewMem(sizeof(KFGHEAD));
  Kfg->language = StrCopy(PLR_C_IMP(Cfg)->language);
  Kfg->TkCnt    = (int)OT_CNT(PLR_C_IMP(Cfg)->Token);
  Kfg->NtCnt    = (int)OT_CNT(PLR_C_IMP(Cfg)->NonTerm);
  Kfg->StartCnt = (int)OT_CNT(PLR_C_IMP(Cfg)->StartNt);
  BUG_RNG1(KFGSYMCNT+Kfg->StartCnt,MAX_OBJECTS);
  Kfg->SNames   = (string*)NewMem(sizeof(string)*KFGSYMCNT);
  for( i=0; i < Kfg->TkCnt; ++i )
  {
    Kfg->SNames[i] = StrCopy(GET_STRTE(PLR_C_IMP(Cfg)->Token,i));
    MaxSymNamLen = MAX(MaxSymNamLen,strlen(Kfg->SNames[i]));
  }
  for( i = Kfg->TkCnt; i < KFGSYMCNT; ++i )
  {
    Kfg->SNames[i] = StrCopy(GET_STRTE(PLR_C_IMP(Cfg)->NonTerm,i-Kfg->TkCnt));
    MaxSymNamLen = MAX(MaxSymNamLen,strlen(Kfg->SNames[i]));
  }
  Kfg->StartIds = INTTE_ARRAY(PLR_C_IMP(Cfg)->StartNt);
  for( i=0; i < Kfg->StartCnt; ++i ) Kfg->StartIds[i] += Kfg->TkCnt;
  Kfg->TokKind  = INTTE_ARRAY(PLR_C_IMP(Cfg)->TokKind);
  NtEquiv       = (OT_Tab*)NewMem(sizeof(OT_Tab)*Kfg->NtCnt);
  Kfg->NtClass  = (int*)NewMem(sizeof(int)*Kfg->NtCnt);
  Kfg->ErrorNt  = (bool*)NewMem(sizeof(bool)*Kfg->NtCnt);
  for( i=0; i < Kfg->NtCnt; ++ i ) 
  {
    NTCLASS(i)      = i; 
    NTEQUIV(i)      = OT_CREATE_ADT(int);
    OT_T_INS(NTEQUIV(i),i);
    Kfg->ErrorNt[i] = OT_GET(bool,PLR_C_IMP(Cfg)->ErrorNt,i);
  }
  LambdaSet     = BS_create(NONTERMCNT);
  NonTermSet    = BS_create(NONTERMCNT);
  NtRelSet      = BS_create(NONTERMCNT * NONTERMCNT);
  TokenSet      = BS_create(TERMCNT);
  StateTab      = NEW_STATETAB;
  KernTab       = OT_CREATE_ADT(unsigned long);
  GoToKern      = NEW_KLISTTAB;
  for( i=0; i < KFGSYMCNT; ++ i )
  { OL_Lst lst = NEW_INTLIST;

    INS_KLIST(lst,GoToKern);
  }
}

static void fillProdTabs(PLR_Cfg Cfg)
/* Initialisierung der Grammatik-Produktionsdaten, Nonterminal Partition */
{ OT_Tab PIdx; int i, j; bool change = False;

  BUG_RNG1(OT_CNT(PLR_C_IMP(Cfg)->Prod),MAX_OBJECTS);
  ProdCnt    = (int)OT_CNT(PLR_C_IMP(Cfg)->Prod) + Kfg->StartCnt;
  PIdx       = SRT_PRODS(PLR_C_IMP(Cfg)->Prod);
  NonTermTab = (KFGNT*)NewMem(NONTERMCNT*sizeof(KFGNT));
  for( i=0; i < NONTERMCNT; ++i )
  {
    if( i < Kfg->NtCnt )
    { PRDFN pkey; int pos;

      pkey.pid = i;
      BUG_VMIN(( pos = (int)FND_PROD(PLR_C_IMP(Cfg)->Prod,PIdx,&pkey) ),0);
      NonTermTab[i].altcnt = 0;
      NonTermTab[i].fstalt = pos - 1;
      for( j = pos - 1; j < (int)OT_CNT(PLR_C_IMP(Cfg)->Prod); ++j )
        if( GET_PROD(PLR_C_IMP(Cfg)->Prod,GET_INTTE(PIdx,j))->pid == i )
          NonTermTab[i].altcnt += 1;
        else break;
      BUG_VMIN(NonTermTab[i].altcnt,0);
    }
    else
    {
      NonTermTab[i].altcnt = 1;
      NonTermTab[i].fstalt = (int)OT_CNT(PLR_C_IMP(Cfg)->Prod) + (i-Kfg->NtCnt);
    }
    NonTermTab[i].states = NEW_INTTAB;
    NonTermTab[i].first  = BS_create(TERMCNT);
  }
  ProdTab   = (KFGPROD*)NewMem(ProdCnt*sizeof(KFGPROD));
  MaxSymCnt = 1;
  for( i=0; i < ProdCnt; ++i )
  {
    if( i < OT_CNT(PLR_C_IMP(Cfg)->Prod) )
    { PRDFN* prod = GET_PROD(PLR_C_IMP(Cfg)->Prod,GET_INTTE(PIdx,i));

      BUG_RNG0(OT_CNT(prod->psymbol),MAX_OBJECTS);
      ProdTab[i].pname  = StrCopy(prod->pname);
      ProdTab[i].id     = Kfg->TkCnt + prod->pid;
      ProdTab[i].method = prod->method;
      ProdTab[i].symcnt = (int)OT_CNT(prod->psymbol);
      MaxSymCnt = ( ProdTab[i].symcnt > MaxSymCnt ) ?
                  ProdTab[i].symcnt : MaxSymCnt;
      if( ProdTab[i].symcnt )
      { int r0 = GET_PSYM(prod->psymbol,0)->sRow,
            c0 = GET_PSYM(prod->psymbol,0)->sCol;
        ProdTab[i].symbols = (int*)NewMem(ProdTab[i].symcnt*sizeof(int));
        ProdTab[i].symfrms = (int*)NewMem(ProdTab[i].symcnt*sizeof(int));
        for( j=0; j < ProdTab[i].symcnt; ++j )
        {
          ProdTab[i].symbols[j] = GET_PSYM(prod->psymbol,j)->symbol;
          if( ! GET_PSYM(prod->psymbol,j)->istoken )
            ProdTab[i].symbols[j] += Kfg->TkCnt;
          if( j > 0 )
          { int cr = GET_PSYM(prod->psymbol,j)->sRow   - r0,
                cc = GET_PSYM(prod->psymbol,j)->sCol   - c0,
                pr = GET_PSYM(prod->psymbol,j-1)->sRow - r0;
            ProdTab[i].symfrms[j] = ( cr > pr ) 
                                    ? -cc : cc;
          }
          else ProdTab[i].symfrms[j] = 0;
        }
      }
      else ProdTab[i].symbols = ProdTab[i].symfrms = (int*)NULL;
      if( pass1NtPartition(&ProdTab[i]) ) change = True;
    }
    else
    { string StNt;
      ProdTab[i].symbols    = (int*)NewMem(sizeof(int));
      ProdTab[i].symfrms    = (int*)NewMem(sizeof(int));
      ProdTab[i].id         = KFGSYMCNT + (i-(int)OT_CNT(PLR_C_IMP(Cfg)->Prod));
      ProdTab[i].method     = PLR_PRD_DFT;
      ProdTab[i].symcnt     = 1;
      ProdTab[i].symbols[0] = 
        Kfg->StartIds[i-(int)OT_CNT(PLR_C_IMP(Cfg)->Prod)];
      StNt                  = Kfg->SNames[ProdTab[i].symbols[0]];
      ProdTab[i].pname      = (string)NewMem(7+strlen(StNt));
      sprintf(ProdTab[i].pname,"Start_%s",StNt);
      ProdTab[i].symfrms[0] = 0;
    }
  }
  OT_DEL_T(PIdx);
  if( change ) pass2NtPartition();
}

static void fillKernTab(void)
/* Bestimmung der Kernelement-Indizes */
{ unsigned long row, col, key; int i, j;

  for( i=1; i <= ProdCnt; ++i )
  {
    row = ((unsigned long)i) << 16;
    row = row & 0xFFFF0000;
    for( j=1; j <= MaxSymCnt + 1; ++j )
    {
      col = j & 0x0000FFFF;
      key = row | col;
      OT_T_INS(KernTab,key);
    }
  }
}

/* ------------------------ First and Lambda Sets --------------------------- */

static bool FirstSetForSymbols(int SymCnt, int* Symbols)
/* Bestimmung der FIRST-Elemente fuer eine Liste von Grammatik-Symbolen */
/* ( Initialisierung des TokenSet ausserhalb ! )                        */
{ bool LambdaInAll = True; int i;

  for( i=0; i < SymCnt; ++i )
  {
    if( isToken(Symbols[i]) )
    {
      BS_setE(Symbols[i],TokenSet);
      LambdaInAll = False; break;
    }
    else
    {
      BS_union(TokenSet,TokenSet,FIRSTSET(Symbols[i]));
      if( ! BS_member( NTIDX(Symbols[i]),LambdaSet) )
      {
        LambdaInAll = False; break;
      }
    }
  }
  return( LambdaInAll );
}

static void fillFirstSets(void)
/* Bestimmung der FIRST-Elemente fuer alle Nonterminals */
{ bool newElements = True; int i;

  /* FIRST-Elemente fuer die Nonterminals ( Huelle ) */
  while( newElements )
  {
    newElements = False;
    for( i = 0; i < ProdCnt; ++i )
    {
      BS_init(TokenSet);
      if( ProdTab[i].symcnt )
      {
        FirstSetForSymbols(ProdTab[i].symcnt,ProdTab[i].symbols);
        newElements |= 
          addNewMembers
          ( 
            FIRSTSET(ProdTab[i].id),FIRSTSET(ProdTab[i].id),TokenSet
          );
      }
    }
  }
  /* FIRST-Elemente fuer die Startsymbole */
  for( i = KFGSYMCNT; i < NTSYMBOL(NONTERMCNT); ++i )
  { int id = Kfg->StartIds[i-KFGSYMCNT];

    if( BS_member(NTIDX(id),LambdaSet) ) BS_setE(NTACCEPT(i),FIRSTSET(i));
  }
}

static void fillLambdaSet(void)
/* Bestimmung aller Nonterminals, aus denen lambda ableitbar ist */
{ BS_Set NonLambdaProds = BS_create(ProdCnt); 
  bool newElements = False; int i;

  for( i=0; i < ProdCnt; ++i )
  {
    if( ! ProdTab[i].symcnt )
    {
      BS_setE(NTIDX(ProdTab[i].id),LambdaSet); newElements = True;
    }
  }
  while( newElements )
  {
    newElements = False;
    for( i=0; i < ProdCnt; ++i )
    {
      if( ! BS_member(NTIDX(ProdTab[i].id),LambdaSet) &&
          ! BS_member(i,NonLambdaProds)                )
      { int j;

        for( j=0; j < ProdTab[i].symcnt; ++j )
        {
          if( isToken(ProdTab[i].symbols[j]) )
          {
            BS_setE(i,NonLambdaProds); break;
          }
          else
            if( ! BS_member(NTIDX(ProdTab[i].symbols[j]),LambdaSet) ) break;
        }
        if( j == ProdTab[i].symcnt )
        {
          BS_setE(NTIDX(ProdTab[i].id),LambdaSet); newElements = True;
        }
      }
    }
  }
  BS_delS(NonLambdaProds);
}

/* ----------------------------- State Table -------------------------------- */

/* ------------------------- LALR(0) Elements ------------------------------- */

static void fillNtRelSet(void)
/* Bestimmung der Nonterminal-Relation und deren Huelle */
{ int i;

  for( i=0; i < ProdCnt; ++i )
    if( ProdTab[i].symcnt && isNonTerm(ProdTab[i].symbols[0]) )
      BS_setGE
      (
        NtRelSet,NONTERMCNT,
        NTIDX(ProdTab[i].id)+1, NTIDX(ProdTab[i].symbols[0])+1
      );
  BS_closure(NtRelSet,NtRelSet,NONTERMCNT);
}

static void Huelle(PSTATE* state)
/* LALR(0)-Huelle einer Menge von Kernelementen */
{ INT i;

  BS_init(NonTermSet);
  for( i=0; i < OL_CNT(state->kern); ++i )
  { INT curr = ( ! i ) ? FST_INTLE(state->kern) : NXT_INTLE(state->kern);
    int idx = PRODUCTION(curr) - 1, pos = POSITION(curr) - 1;

    if( pos != MaxSymCnt && isNonTerm(ProdTab[idx].symbols[pos]) )
    { int row = NTIDX(ProdTab[idx].symbols[pos])+1, j;

      for( j=1; j <= NONTERMCNT; ++j )
        if( BS_member(BS_RELEL(row,j,NONTERMCNT)-1,NtRelSet) )
          BS_setE(j-1,NonTermSet);
    }
  }
}

static void Sprung(OL_Lst Von)
/* Tabelle der Liste von Folge-LALR(0)-Kernen, */
/* ausgehend von den Kernelementen in Von und  */
/* den Huellenelementen in NonTermSet          */
{ OL_Lst Nach; INT i, j;

  for( i=0; i < OL_CNT(Von); ++i )
  { INT curr = ( ! i ) ? FST_INTLE(Von) : NXT_INTLE(Von);
    int idx = PRODUCTION(curr) - 1, pos = POSITION(curr) - 1;

    if( pos != MaxSymCnt )
    {
      Nach = GET_KLIST(GoToKern,ProdTab[idx].symbols[pos]);
      INSS_INTLE(nextProdElement(curr),Nach);
    }
  }
  for( i=0; i < NONTERMCNT; ++i )
  {
    if( BS_member(i,NonTermSet) )
    {
      for( j = FIRSTNTPROD(NTSYMBOL(i)); j <= LASTNTPROD(NTSYMBOL(i)); ++j )
      {
        if( ProdTab[j].symcnt )
        {
          Nach = GET_KLIST(GoToKern,ProdTab[j].symbols[0]);
          INSS_INTLE(nextProdElement(KERNELEMENT(j,1)),Nach);
        }
      }
    }
  }
}

static void setLambdaProdElements(PSTATE* curr)
/* Hinzufuegen der Lambdaproduktionen zu den LALR(0)-Elementen in curr */
{ int i, j;

  for( i=0; i < NONTERMCNT; ++i )
  {
    if( BS_member(i,NonTermSet) )
    {
      for( j = FIRSTNTPROD(NTSYMBOL(i)); j <= LASTNTPROD(NTSYMBOL(i)); ++j )
      {
        if( ! ProdTab[j].symcnt )
        { INT element = KERNELEMENT(j,MaxSymCnt+1),
              pos     = FND_INTLE(element,curr->kern);

          if( ! pos || element != CURR_INTLE(curr->kern) )
          { BS_Set lahead = BS_create(TERMCNT);

            if( ! pos )
            {
              INST_INTLE(element,curr->kern); INS_LOOK(lahead,curr->looks);
            }
            else
            {
              INSC_INTLE(element,curr->kern); INSN_LOOK(lahead,curr->looks,pos);
            }
          }
        }
      }
    }
  }
}

static void setInitialStates(void)
/* Bestimmung der Startzustaende */
{ PSTATE newState; int i;

  for( i=0; i < Kfg->StartCnt; ++i )
  { INT element = KERNELEMENT(ProdCnt-Kfg->StartCnt+i,1);
    BS_Set lahead = BS_create(TERMCNT);

    newState.symbol = -1;
    newState.looks  = NEW_LOOKTAB;
    BS_setE(ACCEPTTK(i),lahead);
    INS_LOOK(lahead, newState.looks);
    newState.kern   = NEW_INTLIST;
    newState.states = NEW_INTLIST;
    INST_INTLE(element, newState.kern);
    INS_STATE(&newState,StateTab);
  }
}

static bool existsFollowState(PSTATE* curr, OL_Lst newKLst)
/* Der Parserzustand ( LALR(0)-Kernelement-Menge ) ist bereits vorhanden */
{ INT i;

  for( i=0; i < OT_CNT(StateTab); ++i )
  {
    if( OL_EQUAL(GET_STATE(StateTab,i)->kern,newKLst) )
    { INT pos = FND_INTLE(i,curr->states);

      if( ! pos ) INST_INTLE(i,curr->states);
      else
      {
        if( i != CURR_INTLE(curr->states) ) INSC_INTLE(i,curr->states);
      }
      return( True );
    }
  }
  return( False );
}

static void setStates(void)
/* Bestimmung der Parserzustaende und der Mengen von LALR(0)-Kernelementen */
{ PSTATE* curr; PSTATE newState; INT i; int j;

  setInitialStates();
  for( i=0; i < OT_CNT(StateTab); ++i )
  {
    curr = GET_STATE(StateTab,i);
    Huelle(curr); Sprung(curr->kern);
    for( j=0; j < KFGSYMCNT; ++j )
    {
      if( OL_CNT(GET_KLIST(GoToKern,j)) )
      {
        if( ! existsFollowState(curr,GET_KLIST(GoToKern,j)) )
        { INT StateId, k;

          newState.symbol = j;
          newState.looks  = NEW_LOOKTAB;
          newState.states = NEW_INTLIST;
          newState.kern   = NEW_INTLIST;
          OL_COPY_L(newState.kern,GET_KLIST(GoToKern,j));
          OL_INIT(GET_KLIST(GoToKern,j));
          for( k=0; k < OL_CNT(newState.kern); ++k )
          { BS_Set lahead = BS_create(TERMCNT);

            INS_LOOK(lahead,newState.looks);
          }
          INS_STATE(&newState,StateTab);
          StateId = OT_CNT(StateTab) - 1;
          INST_INTLE(StateId,curr->states);
          if( isNonTerm(j) ) 
            INS_INTTE((int)StateId,NonTermTab[NTIDX(j)].states);
        }
        else OL_CLEAR(GET_KLIST(GoToKern,j));
      }
    }
  }
  for( i=0; i < OT_CNT(StateTab); ++i )
  {
    curr = GET_STATE(StateTab,i);
    Huelle(curr); setLambdaProdElements(curr);
  }
}

/* ------------------------- LALR(1) Elements ------------------------------- */

static void createProdLaheads(void)
/* Erzeugen der Lookahead-Sets der Huellen-Elemente ( Produktionen ) */
{ int i;

  PLaheads = (BS_Set*)NewMem(sizeof(BS_Set)*ProdCnt);
  for( i=0; i < ProdCnt; ++i ) PLaheads[i] = BS_create(TERMCNT);
}

static void initProdLaheads(void)
/* ( Re- ) Initialisierung der aktuellen Lookahead-Sets der Huellen-Elemente */
{ int i;

  for( i=0; i < ProdCnt; ++i ) BS_init(PLaheads[i]);
}

static void delProdLaheads(void)
/* Loeschen der Lookahead-Sets der Huellen-Elemente */
{ int i;

  for( i=0; i < ProdCnt; ++i ) BS_delS(PLaheads[i]);
  FreeMem(PLaheads);
}

static void LaheadHuelle(INT element, BS_Set lahead)
/* Lookahead-Huelle fuer das Kernelement element */
{ int idx = PRODUCTION(element) - 1, pos = POSITION(element) - 1;
  bool newElements = True;

  if( pos != MaxSymCnt && isNonTerm(ProdTab[idx].symbols[pos]) )
  /* Fuer ein Reduktionselement sind die zulaessigen Folgesymbole */
  /* bereits vollstaendig bestimmt.                               */
  /* Nur ein "Nonterminal"-Kernelement kann fuer die - indirekt - */
  /* zu diesem Nonterminal gehoerenden Reduktionselemente         */
  /* zulaessige Folgesymbole liefern.                             */
  { int i, j;

    BS_init(TokenSet);
    if( pos + 1 < ProdTab[idx].symcnt )
    {
      if( FirstSetForSymbols
          (ProdTab[idx].symcnt-(pos+1),&ProdTab[idx].symbols[pos+1]) )
        BS_union(TokenSet,TokenSet,lahead);
    }
    else BS_union(TokenSet,TokenSet,lahead);
    for( i = 0; i < NONTERMCNT; ++i )
      if( BS_member(i,NonTermSet) && NTSYMBOL(i) == ProdTab[idx].symbols[pos] )
        for( j = FIRSTNTPROD(NTSYMBOL(i)); j <= LASTNTPROD(NTSYMBOL(i)); ++j )
          BS_union(PLaheads[j],PLaheads[j],TokenSet);

    /* Vervollstaendigung der Lookahead-Huelle fuer das aktuelle Kernelement */
    while( newElements )
    {
      newElements = False;
      for( i=0; i < ProdCnt; ++i )
      {
        if( ! BS_empty(PLaheads[i]) && ProdTab[i].symcnt )
        {
          BS_init(TokenSet);
          if( ProdTab[i].symcnt > 1 )
          {
            if( FirstSetForSymbols(ProdTab[i].symcnt-1,&ProdTab[i].symbols[1]) )
              BS_union(TokenSet,TokenSet,PLaheads[i]);
          }
          else BS_union(TokenSet,TokenSet,PLaheads[i]);
          for( j=0; j < ProdCnt; ++j )
            if( ProdTab[i].symbols[0] == ProdTab[j].id          &&
                addNewMembers(PLaheads[j],PLaheads[j],TokenSet)  )
              newElements = True;
        }
      }
    }
  }
}

static bool setFollowLaheads(PSTATE* curr)
/* Bestimmung der Lookahead-Elemente fuer jeden Folgezustand von curr */
{ bool newElements = False; INT i;

  for( i=0; i < OL_CNT(curr->kern); ++i )
  { INT element = ( ! i ) ? FST_INTLE(curr->kern) : NXT_INTLE(curr->kern);
    BS_Set lahead = GET_LOOK(curr->looks,i);

    if( POSITION(element) != MaxSymCnt+1 )
    { OL_Lst readLst = NEW_INTLIST; INT j;

      element = nextProdElement(element);
      for( j=0; j < OL_CNT(curr->states); ++j )
      { INT FState = ( ! j ) ? 
                     FST_INTLE(curr->states) : NXT_INTLE(curr->states);
        PSTATE* follow = GET_STATE(StateTab,FState);
        BS_Set FSet; INT pos, k;

        OL_COPY_L(readLst,follow->kern);
        pos = FND_INTLE(element,readLst);
        if( pos && element == CURR_INTLE(readLst) )
        {
          FSet = GET_LOOK(follow->looks,pos-1);
          if( addNewMembers(FSet,FSet,lahead) )
          {
            BS_setE(FState,StateSet); newElements = True;
          }
        }
        for( k=0; k < OL_CNT(readLst); ++k )
        { INT FElement = ( ! k ) ? FST_INTLE(readLst) : NXT_INTLE(readLst);
          int idx = PRODUCTION(FElement) - 1;

          if( ProdTab[idx].symcnt )
          { INT prevE = prevProdElement(FElement);

            if( ! BS_empty(PLaheads[idx]) && POSITION(prevE) == 1 &&
                ProdTab[idx].symbols[0] == follow->symbol          )
            {
              FSet = GET_LOOK(follow->looks,k);
              if( addNewMembers(FSet,FSet,PLaheads[idx]) )
              {
                BS_setE(FState,StateSet); newElements = True;
              }
            }
          }
        }
      }
      OL_DEL_C(readLst);
    }
  }
  return( newElements );
}

static void setLambdaProdLaheads(PSTATE* curr)
/* Bestimmung der Lookahead-Elemente fuer die Lambdaproduktionen in curr */
{ int i;

  for( i=0; i < ProdCnt; ++i )
  {
    if( ! ProdTab[i].symcnt && ! BS_empty(PLaheads[i]) )
    { INT element = KERNELEMENT(i,MaxSymCnt+1),
          pos     = FND_INTLE(element,curr->kern);

      if( pos && element == CURR_INTLE(curr->kern) )
      { BS_Set lahead = GET_LOOK(curr->looks,pos-1);

        addNewMembers(lahead,lahead,PLaheads[i]);
      }
    }
  }
}

static bool setLaheads(PSTATE* curr)
/* Bestimmung der von curr spontan generierten und sich */
/* ausbreitenden Lookahead-Elemente                     */
{ INT i;

  Huelle(curr); initProdLaheads();
  for( i=0; i < OL_CNT(curr->kern); ++i )
  { INT element = ( ! i ) ? FST_INTLE(curr->kern) : NXT_INTLE(curr->kern);

    LaheadHuelle(element,GET_LOOK(curr->looks,i));
  }
  setLambdaProdLaheads(curr);
  return( setFollowLaheads(curr) );
}

static void fillStateTab(void)
/* Bestimmung der Parserzustaende und der Mengen von LALR(1)-Kernelementen */
{ bool newElements = True; INT i;

  if( profile ) fprintf(STDERR,"creating LALR(0)-Elements ...\n");
  setStates();
  BUG_RNG1(OT_CNT(StateTab),MAX_OBJECTS);
  if( profile ) fprintf(STDERR,"creating LALR(1)-Elements ...\n");
  StateSet = BS_create(OT_CNT(StateTab));
  createProdLaheads();
  for( i=0; i < Kfg->StartCnt; ++i ) BS_setE(i,StateSet);
  while( newElements )
  {
    newElements = False;
    for( i=0; i < OT_CNT(StateTab); ++i )
    {
      if( BS_member(i,StateSet) )
      {
        BS_delE(i,StateSet);
        if( setLaheads(GET_STATE(StateTab,i)) ) newElements = True;
      }
    }
  }
  delProdLaheads();
}

/* --------------------------- Parse Tables --------------------------------- */

static INT setStateShifts(INT idx, OT_Tab shifts)
/* Shift-FolgeZustaende fuer 'StateTab[idx]' */
{ OL_Lst StateLst; INT State, i;

  StateLst = GET_STATE(StateTab,idx)->states;
  for( i=0; i < OL_CNT(StateLst); ++i )
  {
    State = ( ! i ) ? FST_INTLE(StateLst) : NXT_INTLE(StateLst);
    if( isToken( GET_STATE(StateTab,State)->symbol ) )
      INS_INTTE((-(State+1)),shifts);
  }
  return( OT_CNT(shifts) );
}

static INT findStateShifts(OT_Tab shifts, OT_Tab MstShift, OT_Tab DtlShift)
/* finden des Zustands mit identischen Shift-FolgeZustaenden ( oder -1 ) */
{ INT cnt = OT_CNT(MstShift), i, j, k;
  
  for( i=0; i < cnt; ++i )
  { int dcnt = 0;

    if( ( k = j = ( GET_INTTE(MstShift,i) - 1 ) ) >= 0 )
    {
      for( ; j >= 0; --j )
        if( GET_INTTE(DtlShift,j) < 0 ) ++dcnt;
        else                            break;
      if( ( j = OT_CNT(shifts) ) == dcnt )
      { bool equality = True;

        for( j -= 1; j >= 0; --j, --k )
          if( GET_INTTE(DtlShift,k) != GET_INTTE(shifts,j) )
          {
            equality = False; break;
          }
        if( equality ) return( i );
      }
    }
  }
  return( -1 );
}

static void fillShiftTabs(OT_Tab MstShift, OT_Tab DtlShift)
/* Bestimmung der Shift-Tabellen                                            */
/* MstShift : 1 Eintrag / Zustand --- DtlShift-Index des 1. Symbols oder -1 */
/* DtlShift : Liste von Eintraegen fuer jeden Zustand mit Shift-Aktionen    */
/*            ( -(StateIdx1+1), ..., -(StateIdxN+1) ,TokenN, ..., Token1 )  */
{ OT_Tab cshifts = NEW_INTTAB; INT cnt, i, j;

  for( i=0; i < OT_CNT(StateTab); ++i )
  { int lalrel = -1;

    if( ! ( cnt = setStateShifts(i,cshifts) ) ) /* keine Shift-Folgezustaende */
      INS_INTTE(lalrel,MstShift);
    else
    { INT idx = findStateShifts(cshifts,MstShift,DtlShift);

      if( idx >= 0 ) /* vorhandene Shift-Folgezustaende */
        INS_INTTE(GET_INTTE(MstShift,idx),MstShift);
      else /* neue Shift-Folgezustaende */
      { 
        lalrel = (int)(OT_CNT(DtlShift) + cnt); INS_INTTE(lalrel,MstShift);
        for( j=0; j < cnt; ++j ) INS_INTTE(GET_INTTE(cshifts,j),DtlShift);
        for( j=cnt-1; j >= 0; --j )
        {
          lalrel = GET_STATE(StateTab,ABS((GET_INTTE(cshifts,j)+1)))->symbol;
          INS_INTTE(lalrel,DtlShift);
        }
      }
    }
    OT_CLEAR(cshifts);
  }
  OT_DEL_T(cshifts);
}

static void setStateGoTos(INT idx, OT_Tab GoToDft, bool sign)
/* GoTo-QuellZustaende fuer Zielzustand 'StateTab[idx]' */
{ OL_Lst StateLst; INT State, cnt = OT_CNT(StateTab), i, j;

  for( i=0; i < cnt; ++i )
  {
    StateLst = GET_STATE(StateTab,i)->states;
    for( j=0; j < OL_CNT(StateLst); ++j )
    {
      State = ( ! j ) ? FST_INTLE(StateLst) : NXT_INTLE(StateLst);
      if( State == idx ) 
        INS_INTTE(sign?(int)(-(i+1)):(int)(i+1),GoToDft);
    }
  }
}

static void fillGoToTabs(OT_Tab GoTo, OT_Tab GoToDft)
/* Bestimmung der Sprungtabellen                                              */
/* GoTo    : 1 Eintrag / Nonterminal -- Index des 1. GoToDft-Eintrags         */
/* GoToDft : Folge von Eintraegen der Art                                     */
/* (old)     'toStateIdx+1+StateCnt, [ fromStateIdx+1 ... | 0 ( Default ) ]'  */
/* (new)     'toStateIdx+1+StateCnt, [ [- (Default) ] (fromStateIdx+1) ... ]' */
{ int lalrel, i;

  for( i=0; i < NONTERMCNT && isNonTerm(NTSYMBOL(i)); ++i )
  { INT cnt = OT_CNT(NonTermTab[i].states), j;

    INS_INTTE((int)OT_CNT(GoToDft),GoTo);
    if( cnt <= 1 )
    {
      if( cnt == 1 ) /* unique StateSymbol */
      {
        lalrel = (int)(OT_CNT(StateTab)+GET_INTTE(NonTermTab[i].states,0)+1); 
        INS_INTTE(lalrel,GoToDft); 
        setStateGoTos(GET_INTTE(NonTermTab[i].states,0),GoToDft,True);
      }
      else /* unreacheable NonTerminal */
      {
        fprintf
        (
          STDERR,"Warning: unreacheable Nonterminal '%s'\n",
          Kfg->SNames[NTSYMBOL(i)]
        );
        INS_INTTE(0,GoToDft);
      }
    }
    else /* more than one State with StateSymbol */
    { 
      for( j=0; j < cnt; ++j )
      { 
        lalrel = (int)(OT_CNT(StateTab)+GET_INTTE(NonTermTab[i].states,j)+1); 
        INS_INTTE(lalrel,GoToDft);
        setStateGoTos(GET_INTTE(NonTermTab[i].states,j),GoToDft,False);
      }
    }
  }
}

static int nextRedElement(PSTATE* state, int* RedElement, int idx)
/* Position der/s ersten/naechsten Reduktionselements des Zustands oder 0 */
/* RedElement enthaelt nach Aufruf das Reduktionselement ( ProdIdx + 1 )  */
{ INT i;

  for( i=idx; i < OL_CNT(state->kern); ++i )
  { INT element = ( ! i ) ? FST_INTLE(state->kern) : NXT_INTLE(state->kern);

    if( POSITION(element) == MaxSymCnt + 1 )
      return( ( *RedElement = PRODUCTION(element) ), (int)(i+1) );
  }
  return( 0 );
}

static int RedAction(int RedElement)
/* Reduktionselement ( Index der Produktion + 1 ) oder ACT_ACCEPT */
{
  return( ( isNonTerm(ProdTab[RedElement-1].id) ) ? RedElement : ACT_ACCEPT );
}

static void fillActExpLst
            (
              OT_Tab ActExp, int RedElement, BS_Set lahead, bool negFlg
            )
/* Bestimmung der Ausnahme-Aktionen fuer das Reduktionselement */
/* Alternierende +/- Folge von Eintraegen der Art              */
/* 'PROD_OFFSET + Red.Nr ( >= 1 ), Token ( >= 1 ) ...' fuer    */
/* jeden Zustand mit mehreren Reduktionselementen              */
{ int lalrel = RedAction(RedElement),
      i      = ( lalrel != ACT_ACCEPT ) ? 0 : Kfg->TkCnt;

  lalrel += PROD_OFFSET;
  lalrel  = ( negFlg ) ? -lalrel : lalrel; INS_INTTE(lalrel,ActExp);
  for( ; i < TERMCNT; ++i )
  {
    if( BS_member(i,lahead) )
    {
      lalrel = ( negFlg ) ? -(i+1) : i+1; INS_INTTE(lalrel,ActExp);
    }
  }
}

static void fillActTabs(OT_Tab ActDft, OT_Tab ActExp)
/* Bestimmung der Default/Ausnahme-Aktionstabellen                          */
/* ActDft : 1 Eintrag / Zustand -- ACT_ACCEPT, ACT_ERROR, Red.Nr ( >= 1 ),  */
/*                                 negativer Index des 1. ActExp-Eintrags   */
/* ActExp : siehe fillActExpLst                                             */
{ INT i;

  for( i=0; i < OT_CNT(StateTab); ++i )
  { PSTATE* curr = GET_STATE(StateTab,i);
    int lalrel = ACT_ERROR, FstRedElement, RedElement, pos;
    int FstPos = nextRedElement(curr,&FstRedElement,0);

    if( ! FstPos ) /* Default-Aktion: ACT_ERROR */
      INS_INTTE(lalrel,ActDft);
    else
    {
      pos = nextRedElement(curr,&RedElement,FstPos);
      if( ! pos ) /* Default-Aktion: ACT_ACCEPT/Reduktion */
      { /* veraenderte Bedingung fuer Fehlerbehandlung */
        lalrel = RedAction(FstRedElement); INS_INTTE(lalrel,ActDft);
      }
      else /* Ausnahme-Aktionen: ACT_ACCEPT/Reduktion */
      { bool negFlg = ( ! OT_CNT(ActExp)                       || 
                        GET_INTTE(ActExp,OT_CNT(ActExp)-1) < 0  ) ?
                        False : True;

        lalrel = (int)(-OT_CNT(ActExp)); INS_INTTE(lalrel,ActDft);
        fillActExpLst
        (
          ActExp,FstRedElement,GET_LOOK(curr->looks,FstPos-1),negFlg
        );
        while( pos > 0 )
        {
          fillActExpLst(ActExp,RedElement,GET_LOOK(curr->looks,pos-1),negFlg);
          pos = nextRedElement(curr,&RedElement,pos);
        }
      }
    }
  }
}

static PLR_Tab fillParseTab(void)
/* Bestimmung der Analyse-Tabellen der Grammatik */
{ PARSETAB* PTab  = (PARSETAB*)NewMem(sizeof(PARSETAB));
  OT_Tab MstShift = NEW_INTTAB;
  OT_Tab DtlShift = NEW_INTTAB;
  OT_Tab GoTo     = NEW_INTTAB;
  OT_Tab GoToDft  = NEW_INTTAB;
  OT_Tab ActDft   = NEW_INTTAB;
  OT_Tab ActExp   = NEW_INTTAB;

  PTab->Kfg     = Kfg;
  PTab->ProdCnt = ProdCnt;
  PTab->Prod    = ProdTab;

  fillShiftTabs(MstShift,DtlShift);
  BUG_VMIN(PTab->tabs.MstShiftCnt = (int)OT_CNT(MstShift),0);
  PTab->tabs.MstShift    = INTTE_ARRAY(MstShift);
  OT_DEL_T(MstShift);
  PTab->tabs.DtlShiftCnt = (int)OT_CNT(DtlShift);
  PTab->tabs.DtlShift    = INTTE_ARRAY(DtlShift);
  OT_DEL_T(DtlShift);

  fillGoToTabs(GoTo,GoToDft);
  BUG_VMIN(PTab->tabs.GoToCnt = (int)OT_CNT(GoTo),0);
  PTab->tabs.GoTo       = INTTE_ARRAY(GoTo);
  OT_DEL_T(GoTo);
  BUG_VMIN(PTab->tabs.GoToDftCnt = (int)OT_CNT(GoToDft),0);
  PTab->tabs.GoToDft    = INTTE_ARRAY(GoToDft);
  OT_DEL_T(GoToDft);

  fillActTabs(ActDft,ActExp);
  BUG_VMIN(PTab->tabs.ActDftCnt = (int)OT_CNT(ActDft),0);
  PTab->tabs.ActDft    = INTTE_ARRAY(ActDft);
  OT_DEL_T(ActDft);
  PTab->tabs.ActExpCnt = (int)OT_CNT(ActExp);
  PTab->tabs.ActExp    = INTTE_ARRAY(ActExp);
  OT_DEL_T(ActExp);

  return( PTab );
}

/* ---------------------------- Printing ------------------------------------ */

static int indent = 2;

static void printINTObj(StdCPtr obj)
/* prints an INT-object */
{
  fprintf(STDOUT,"%ld",(long)obj);
}

static void printTokSet(BS_Set s)
/* prints a token set */
{ int i;

  fprintf(STDOUT,"{");
  for( i=0; i < TERMCNT; ++i )
    if( BS_member(i,s) )
      fprintf(STDOUT," \"%s\" ", i<Kfg->TkCnt? Kfg->SNames[i] : "Accept");
  fprintf(STDOUT,"}");
}

static void printFirstNtSets(void)
/* prints first sets */
{ int i;

  NL; fprintf(STDOUT,"--- First Sets ---"); NL; NL;
  for( i=0; i < NONTERMCNT; ++i )
  {
    fprintf
    (
      STDOUT,"First %s%s = ",
      i<Kfg->NtCnt? Kfg->SNames[NTSYMBOL(i)] : "Start",
      BS_member(i,LambdaSet) ? "(lambda)" : ""
    );
    printTokSet(FIRSTSET(NTSYMBOL(i))); NL;
  }
}

static void ShiftConflicts(BS_Set cl,OL_Lst states,bool printing)
/* prints shift-reduce conflicts for current kern element */
{ bool flag = False; INT i;

  for( i=0; i < OL_CNT(states); ++i )
  { INT fs = ( ! i ) ? FST_INTLE(states) : NXT_INTLE(states);
    int symbol = GET_STATE(StateTab,fs)->symbol;

    if( isToken(symbol) && BS_member(symbol,cl) )
    {
      if( printing )
      {
        NLIND;
        fprintf
        (
          STDOUT, "Shift-Reduce-Konflikt mit Shift '%s' zu Zustand: %ld",
          Kfg->SNames[symbol], (long)fs
        );
        flag = True;
      }
      ++conflictCnt;
    }
  }
  if( flag ) NL;
}

static void RedConflicts
            (
              int clidx,BS_Set cl,OT_Tab looks,OL_Lst readLst,bool printing
            )
/* prints reduce-reduce conflicts for current kern element */
{ bool flag = False; INT i;

  for( i = clidx + 1; i < OL_CNT(readLst); ++i )
  { INT fk = NXT_INTLE(readLst); int pos = POSITION(fk);

    if( pos == MaxSymCnt + 1                                &&
        ! BS_empty(BS_inter(TokenSet,cl,GET_LOOK(looks,i)))  )
    {
      if( printing )
      { int idx = PRODUCTION(fk) - 1, id = ProdTab[idx].id; char c = ' ';

        if( id >= KFGSYMCNT )
        {
          id = Kfg->StartIds[id-KFGSYMCNT]; c  = '\'';
        }
        NLIND;
        fprintf
        (
          STDOUT,"Reduce-Reduce-Konflikt mit Element: %s%c ( %3d )",
          Kfg->SNames[id], c, pos
        );
        flag = True;
      }
      ++conflictCnt;
    }
  }
  if( flag ) NL;
}

static void printStateTab(void)
/* prints state table and conflicts */
{ OL_Lst readLst = NEW_INTLIST; INT i, k; int cnt;

  NL; fprintf(STDOUT,"--- State Table and Conflicts ---"); NL; NL;
  for( i=0; i < OT_CNT(StateTab); ++i )
  { PSTATE* curr = GET_STATE(StateTab,i);

    fprintf
    (
      STDOUT,"State: %3d / Symbol: %s", 
      (int)i, ( curr->symbol == -1 ) ? "None" : Kfg->SNames[curr->symbol]
    );
    NL; NLIND;
    fprintf(STDOUT,"Follow States:");
    PRT_INTLES(curr->states,indent+2);
    NL; NLIND;
    fprintf(STDOUT,"LALR(1) - Elements:");
    for( k=0; k < OL_CNT(curr->kern); ++k )
    { INT ck = ( ! k ) ? FST_INTLE(curr->kern) : NXT_INTLE(curr->kern);
      int idx = PRODUCTION(ck) - 1, pos = POSITION(ck), id = ProdTab[idx].id, j;
      BS_Set cl = GET_LOOK(curr->looks,k); char c = ' ';

      OL_COPY_L(readLst,curr->kern);
      if( id >= KFGSYMCNT )
      {
        id = Kfg->StartIds[id-KFGSYMCNT]; c = '\'';
      }
      TAB(indent+2);
      fprintf(STDOUT,"%s%c ( %3d ) :: ",Kfg->SNames[id],c,pos);
      if( c != '\'' )
      { int fstcol = indent + 2 + strlen(Kfg->SNames[id]) + 13;

        for( j=0; j < ProdTab[idx].symcnt; ++j )
        {
          if( ! ( j % MAX_PR_SCOLS ) && j ) TAB(fstcol);
          if( pos == j+1 ) fprintf(STDOUT,"^ ");
          fprintf(STDOUT,"%s ",Kfg->SNames[ProdTab[idx].symbols[j]]);
        }
        if( pos == MaxSymCnt + 1 ) fprintf(STDOUT,"^");
      }
      else
      {
        if( pos == 1 ) fprintf(STDOUT,"^ ");
        fprintf(STDOUT,"%s ",Kfg->SNames[id]);
        if( pos == MaxSymCnt + 1 ) fprintf(STDOUT,"^");
      }
      TAB(indent+2);
      fprintf(STDOUT,"Lookaheads: ");
      for( j=0, cnt=0; j < TERMCNT; ++j )
      {
        if( BS_member(j,cl) )
        {
          if( ! ( cnt % MAX_PR_SCOLS ) && cnt ) TAB(indent+14);
          if( j < Kfg->TkCnt ) fprintf(STDOUT,"%s ",Kfg->SNames[j]);
          else                 fprintf(STDOUT,"Accept[%d] ",TKACCEPT(j));
          ++cnt;
        }
      }
      NL;
      if( pos == MaxSymCnt + 1 )
      {
        RedConflicts((int)k,cl,curr->looks,readLst,True);
        ShiftConflicts(cl,curr->states,True);
      }
    }
    NL;
  }
  OL_DEL_C(readLst);
}

static void countConflicts(void)
/* counts conflicts */
{ OL_Lst readLst = NEW_INTLIST; INT i, k;

  for( i=0; i < OT_CNT(StateTab); ++i )
  { PSTATE* curr = GET_STATE(StateTab,i);

    for( k=0; k < OL_CNT(curr->kern); ++k )
    { INT ck = ( ! k ) ? FST_INTLE(curr->kern) : NXT_INTLE(curr->kern);
      int pos = POSITION(ck); BS_Set cl = GET_LOOK(curr->looks,k);

      OL_COPY_L(readLst,curr->kern);
      if( pos == MaxSymCnt + 1 )
      {
        RedConflicts((int)k,cl,curr->looks,readLst,False);
        ShiftConflicts(cl,curr->states,False);
      }
    }
  }
  OL_DEL_C(readLst);
}

static void printNtClasses(void)
/* prints nonterminal classes */
{ int i;
  NL; fprintf(STDOUT,"--- Nonterminal Classes ---"); NL; NL;
  for( i=0; i < Kfg->NtCnt; ++i )
  {
    fprintf
    (
      STDOUT,"%*s |--> %s",
      MaxSymNamLen,Kfg->SNames[NTSYMBOL(i)],
      Kfg->SNames[NTSYMBOL(NTCLASS(i))]
    );
    NL;
  }
}

/*====================== Parser Generator Interface ==========================*/

/*I----------------------- Grammar definition ------------------------------- */

PLR_Cfg PLR_createCfg(string Language)
/* creates a context free grammar definition named 'Language' */
{ KFGDFN* cfg;

  BUG_NULL(Language);
  cfg           = (KFGDFN*)NewMem(sizeof(KFGDFN));
  cfg->language = StrCopy(Language);
  cfg->Token    = NEW_STRTAB;
  cfg->TkIdx    = (OT_Tab)NULL;
  cfg->NonTerm  = NEW_STRTAB;
  cfg->NtIdx    = (OT_Tab)NULL;
  cfg->StartNt  = NEW_INTTAB;
  cfg->TokKind  = NEW_INTTAB;
  cfg->ErrorNt  = OT_CREATE_ADT(bool);
  cfg->Prod     = NEW_PRODTAB;
  return( cfg );
}

int PLR_addTK(PLR_Cfg Cfg, string Token, int kind)
/* adds token 'Token' of type 'kind' 
   ( token or keyword, see [cfg_dfn] )
   to definition 'Cfg' 
*/
{
  BUG_NULL(Cfg); BUG_NULL(Token);
  BUG_VEQU(PLR_C_IMP(Cfg)->TkIdx,(OT_Tab)NULL);
  BUG_VEQU(CFG_TERMINAL(kind),True);
  INS_INTTE(kind,PLR_C_IMP(Cfg)->TokKind);
  return( (int)INS_STRTE(Token,PLR_C_IMP(Cfg)->Token) );
}

int PLR_addNT(PLR_Cfg Cfg, string NonTerm, bool catchError)
/* adds nonterminal 'NonTerm' to definition 'Cfg' 
   catchError --> use 'NonTerm' as reparse point
*/
{
  BUG_NULL(Cfg); BUG_NULL(NonTerm);
  BUG_VEQU(PLR_C_IMP(Cfg)->NtIdx,(OT_Tab)NULL);
  OT_PUSH(catchError,PLR_C_IMP(Cfg)->ErrorNt);
  return( (int)INS_STRTE(NonTerm,PLR_C_IMP(Cfg)->NonTerm) );
}

void PLR_endSD(PLR_Cfg Cfg)
/* symbol definition end; 
   completes token and nonterminal definition
*/
{
  BUG_NULL(Cfg);
  BUG_RNG0(OT_CNT(PLR_C_IMP(Cfg)->Token),MAX_OBJECTS);
  BUG_RNG1(OT_CNT(PLR_C_IMP(Cfg)->NonTerm),MAX_OBJECTS);
  PLR_C_IMP(Cfg)->TkIdx = SRT_STRTES(PLR_C_IMP(Cfg)->Token);
  PLR_C_IMP(Cfg)->NtIdx = SRT_STRTES(PLR_C_IMP(Cfg)->NonTerm);
}

int PLR_addSN(PLR_Cfg Cfg, string StartNt)
/* adds startsymbol 'StartNt' to definition 'Cfg' */
{ INT nt;

  BUG_NULL(Cfg); BUG_NULL(StartNt);
  BUG_VNEQ(PLR_C_IMP(Cfg)->NtIdx,(OT_Tab)NULL);
  nt = FND_STRTE(PLR_C_IMP(Cfg)->NonTerm,PLR_C_IMP(Cfg)->NtIdx,StartNt);
  BUG_VMIN(nt,0);
  nt = GET_INTTE(PLR_C_IMP(Cfg)->NtIdx,nt-1);
  return( (int)INS_INTTE((int)nt,PLR_C_IMP(Cfg)->StartNt) );
}

int PLR_addST(PLR_Cfg Cfg, string SpecTk)
/* marks 'SpecTk' as special comment token */
{ INT tk; int kind;

  BUG_NULL(Cfg); BUG_NULL(SpecTk);
  BUG_VNEQ(PLR_C_IMP(Cfg)->TkIdx,(OT_Tab)NULL);
  tk   = FND_STRTE(PLR_C_IMP(Cfg)->Token,PLR_C_IMP(Cfg)->TkIdx,SpecTk);
  BUG_VMIN(tk,0);
  tk   = GET_INTTE(PLR_C_IMP(Cfg)->TkIdx,tk-1);
  kind = CFG_TERM_TO_COM(GET_INTTE(PLR_C_IMP(Cfg)->TokKind,tk));
  UPD_INTTE(kind,PLR_C_IMP(Cfg)->TokKind,tk);
  return tk;
}

int PLR_addPR(PLR_Cfg Cfg, string PName, int Method, string NonTerm)
/* adds production 'NonTerm'::'PName' with layout hint 'Method' 
   ( default=0, see [prs_dfn] ) to definition 'Cfg' 
*/
{ PRDFN prod;

  BUG_NULL(Cfg); BUG_NULL(PName); BUG_NULL(NonTerm);
  BUG_RNG0(Method,PLR_PRD_NON);
  BUG_VNEQ(PLR_C_IMP(Cfg)->NtIdx,(OT_Tab)NULL);
  prod.pname   = StrCopy(PName);
  prod.method  = Method;
  prod.pid     = 
    (int)FND_STRTE(PLR_C_IMP(Cfg)->NonTerm,PLR_C_IMP(Cfg)->NtIdx,NonTerm);
  BUG_VMIN(prod.pid,0);
  prod.pid     = GET_INTTE(PLR_C_IMP(Cfg)->NtIdx,prod.pid-1);
  prod.psymbol = NEW_PSYMTAB;
  return( (int)INS_PROD(&prod,PLR_C_IMP(Cfg)->Prod) );
}

int PLR_addPT(PLR_Cfg Cfg, string Token, int sRow, int sCol)
/* adds token 'Token' to current production of definition 'Cfg'; 
   The symbol position 'sRow', 'sCol' is used as layout hint.
*/
{ PSDFN psym; INT idx;

  BUG_NULL(Cfg); BUG_NULL(Token);
  idx          = OT_CNT(PLR_C_IMP(Cfg)->Prod) - 1;
  BUG_VMIN(idx,-1);
  psym.symbol  = 
    (int)FND_STRTE(PLR_C_IMP(Cfg)->Token,PLR_C_IMP(Cfg)->TkIdx,Token);
  BUG_VMIN(psym.symbol,0);
  psym.symbol  = GET_INTTE(PLR_C_IMP(Cfg)->TkIdx,psym.symbol-1);
  psym.istoken = True;
  psym.sRow    = sRow;
  psym.sCol    = sCol;
  return( (int)INS_PSYM(&psym,GET_PROD(PLR_C_IMP(Cfg)->Prod,idx)->psymbol) );
}

int PLR_addPN(PLR_Cfg Cfg, string NonTerm, int sRow, int sCol)
/* adds nonterminal 'NonTerm' to current production of definition 'Cfg'; 
   The symbol position 'sRow', 'sCol' is used as layout hint.
*/
{ PSDFN psym; INT idx;

  BUG_NULL(Cfg); BUG_NULL(NonTerm);
  idx          = OT_CNT(PLR_C_IMP(Cfg)->Prod) - 1;
  BUG_VMIN(idx,-1);
  psym.symbol  =
    (int)FND_STRTE(PLR_C_IMP(Cfg)->NonTerm,PLR_C_IMP(Cfg)->NtIdx,NonTerm);
  BUG_VMIN(psym.symbol,0);
  psym.symbol  = GET_INTTE(PLR_C_IMP(Cfg)->NtIdx,psym.symbol-1);
  psym.istoken = False;
  psym.sRow    = sRow;
  psym.sCol    = sCol;
  return( (int)INS_PSYM(&psym,GET_PROD(PLR_C_IMP(Cfg)->Prod,idx)->psymbol) );
}

void PLR_delCfg(PLR_Cfg Cfg)
/* removes grammar definition 'Cfg' */
{
  BUG_NULL(Cfg);
  FreeMem(PLR_C_IMP(Cfg)->language);
  OT_DEL_T(PLR_C_IMP(Cfg)->Token);
  OT_DEL_T(PLR_C_IMP(Cfg)->TkIdx);
  OT_DEL_T(PLR_C_IMP(Cfg)->NonTerm);
  OT_DEL_T(PLR_C_IMP(Cfg)->NtIdx);
  OT_DEL_T(PLR_C_IMP(Cfg)->StartNt);
  OT_DEL_T(PLR_C_IMP(Cfg)->TokKind);
  OT_DEL_T(PLR_C_IMP(Cfg)->ErrorNt);
  OT_DEL_T(PLR_C_IMP(Cfg)->Prod);
  FreeMem(Cfg);
}

/*I--------------------- Parse table creation ------------------------------- */

PLR_Tab PLR_createTab(PLR_Cfg Cfg, bool verbose, bool diagnose)
/* creates the corresponding parse table for definition 'Cfg'
   verbose --> entertainment
   diagnose --> print conflict / result informations
*/
{ PLR_Tab PTab = (PLR_Tab)NULL;

  BUG_NULL(Cfg);
  profile = verbose;
  if( profile ) fprintf(STDERR,"begin parse table creation ...\n");
  if( profile ) fprintf(STDERR,"reading grammar information ...\n");
  setKfgGlobals(Cfg);
  fillProdTabs(Cfg);
  if( profile ) fprintf(STDERR,"creating element keys ...\n");
  fillKernTab();
  if( profile ) fprintf(STDERR,"creating lambda set ...\n");
  fillLambdaSet();
  if( profile ) fprintf(STDERR,"creating nonterminal relation ...\n");
  fillNtRelSet();
  if( profile ) fprintf(STDERR,"creating first sets ...\n");
  fillFirstSets();
  if( profile ) fprintf(STDERR,"creating states ...\n");
  fillStateTab();
  if( profile ) fprintf(STDERR,"creating parse tables ...\n");
  PTab = fillParseTab();
  if( diagnose )
  {
    fprintf(STDERR,"printing statistics ...\n");
    printFirstNtSets();
    printStateTab();
    printNtClasses();
  }
  else countConflicts();
  if( conflictCnt )
    fprintf(STDERR,"%d conflicts. Use -diagnose !\n",conflictCnt);
  if( profile ) fprintf(STDERR,"deleting ...\n");
  delCreateGlobals();
  if( profile ) fprintf(STDERR,"end parse table creation ...\n");

  return( PTab );
}
